#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <commdlg.h>
#include <richedit.h>
#include "AkelDLL.h"
#include "ConvFunc.h"
#include "StackFunc.h"
#include "Resources\resource.h"


//Include conversion functions
#define xatoiA
#define xatoiW
#include "ConvFunc.h"

//Include stack functions
#define StackGetElement
#define StackInsertAfter
#define StackInsertBefore
#define StackInsertIndex
#define StackDelete
#define StackMoveAfter
#define StackMoveBefore
#define StackMoveIndex
#define StackCopy
#define StackClear
#include "StackFunc.h"

//Defines
#define STRID_SESSION            1
#define STRID_OPEN               2
#define STRID_SAVE               3
#define STRID_COPY               4
#define STRID_RENAME             5
#define STRID_DELETE             6
#define STRID_MENU_OPEN          7
#define STRID_MENU_ACTIVATE      8
#define STRID_MENU_ADD           9
#define STRID_MENU_MOVEDOWN      10
#define STRID_MENU_MOVEUP        11
#define STRID_MENU_DELETE        12
#define STRID_MENU_DELETEOLD     13
#define STRID_NEWNAME            14
#define STRID_SETTINGS           15
#define STRID_OPENONSTART        16
#define STRID_SAVEONEXIT         17
#define STRID_MODALDIALOG        18
#define STRID_OK                 19
#define STRID_CANCEL             20
#define STRID_CLOSE              21
#define STRID_CURRENTSESSION     22
#define STRID_ALREADY_EXIST      23
#define STRID_RENAME_ERROR       24
#define STRID_SESSION_CHANGED    25
#define STRID_CONFIRM_DELETE     26
#define STRID_SDI_ISNTSUPPORTED  27
#define STRID_FILTER             28

#define DLLA_SESSIONS_OPEN 1

#define BUFFER_SIZE 8192

#define OF_MAIN            0x1
#define OF_WINDOWRECT      0x2

#define APP_MDI_CLASSA               "AkelPad MDI Class"
#define APP_MDI_CLASSW              L"AkelPad MDI Class"

//DIALOGRESIZE
#define DRS_SIZE  0x1 //Resize control
#define DRS_MOVE  0x2 //Move control
#define DRS_X     0x4 //X value
#define DRS_Y     0x8 //Y value

//Combobox edit ID
#define IDC_COMBOBOXEDIT       1001

//Dialog messages
#define AKDLL_UPDATESAVEBUTTON  (WM_USER + 1)

typedef struct _SESSIONITEMA {
  struct _SESSIONITEMA *next;
  struct _SESSIONITEMA *prev;
  char szItemFile[MAX_PATH];
  int nTabActive;
  int nCodePage;
  int nSelStart;
  int nSelEnd;
  BOOL bWordWrap;
  BOOL bReadOnly;
  BOOL bOvertypeMode;
} SESSIONITEMA;

typedef struct _SESSIONITEMW {
  struct _SESSIONITEMW *next;
  struct _SESSIONITEMW *prev;
  wchar_t wszItemFile[MAX_PATH];
  int nTabActive;
  int nCodePage;
  int nSelStart;
  int nSelEnd;
  BOOL bWordWrap;
  BOOL bReadOnly;
  BOOL bOvertypeMode;
} SESSIONITEMW;

typedef struct _SESSIONA {
  struct _SESSIONA *next;
  struct _SESSIONA *prev;
  char szSessionName[MAX_PATH];
  HSTACK hItemsStack;
  BOOL bModified;
} SESSIONA;

typedef struct _SESSIONW {
  struct _SESSIONW *next;
  struct _SESSIONW *prev;
  wchar_t wszSessionName[MAX_PATH];
  HSTACK hItemsStack;
  BOOL bModified;
} SESSIONW;

typedef struct {
  HWND *lpWnd;   //Control window
  DWORD dwType;  //See DRS_* defines
} DIALOGRESIZE;

//Functions prototypes
DWORD WINAPI ThreadProcA(LPVOID lpParameter);
DWORD WINAPI ThreadProcW(LPVOID lpParameter);
LRESULT CALLBACK MainDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK MainDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK SettingsDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK SettingsDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK InputBoxDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK InputBoxDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewListBoxProcA(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewListBoxProcW(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewMainProcA(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewMainProcW(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

BOOL LoadSessionFileA(SESSIONA *ss);
BOOL LoadSessionFileW(SESSIONW *ss);
BOOL RenameSessionFileA(char *pOldSessionName, char *pNewSessionName);
BOOL RenameSessionFileW(wchar_t *wpOldSessionName, wchar_t *wpNewSessionName);
void OpenSessionFileA(char *pSessionName);
void OpenSessionFileW(wchar_t *wpSessionName);
void SaveSessionFileA(SESSIONA *ss);
void SaveSessionFileW(SESSIONW *ss);
BOOL DeleteSessionFileA(char *pSessionName);
BOOL DeleteSessionFileW(wchar_t *wpSessionName);
int SaveSessionPromptA();
int SaveSessionPromptW();

void FillSessionsA(HSTACK *hStack);
void FillSessionsW(HSTACK *hStack);
SESSIONA* AddEmptySessionA(HSTACK *hStack, char *pSessionName);
SESSIONW* AddEmptySessionW(HSTACK *hStack, wchar_t *wpSessionName);
SESSIONA* AddCurrentSessionA(HSTACK *hStack, char *pSessionName);
SESSIONW* AddCurrentSessionW(HSTACK *hStack, wchar_t *wpSessionName);
SESSIONA* GetSessionA(HSTACK *hStack, char *pSessionName);
SESSIONW* GetSessionW(HSTACK *hStack, wchar_t *wpSessionName);
void CopySessionA(SESSIONA *ssSource, SESSIONA *ssTarget);
void CopySessionW(SESSIONW *ssSource, SESSIONW *ssTarget);
void OpenSessionA(HSTACK *hStack);
void OpenSessionW(HSTACK *hStack);
void CloseSessionA(HSTACK *hStack);
void CloseSessionW(HSTACK *hStack);
void DeleteSessionA(HSTACK *hStack, SESSIONA *ss);
void DeleteSessionW(HSTACK *hStack, SESSIONW *ss);
void FreeSessionsA(HSTACK *hStack);
void FreeSessionsW(HSTACK *hStack);

SESSIONITEMA* AddSessionItemA(SESSIONA *ss);
SESSIONITEMW* AddSessionItemW(SESSIONW *ss);
SESSIONITEMA* GetSessionItemA(SESSIONA *ss, int nIndex);
SESSIONITEMW* GetSessionItemW(SESSIONW *ss, int nIndex);
void MoveSessionItemA(SESSIONA *ss, int nOldIndex, int nNewIndex);
void MoveSessionItemW(SESSIONW *ss, int nOldIndex, int nNewIndex);
void OpenSessionItemA(SESSIONITEMA *si);
void OpenSessionItemW(SESSIONITEMW *si);
BOOL CloseSessionItemA(SESSIONITEMA *si);
BOOL CloseSessionItemW(SESSIONITEMW *si);
void DeleteSessionItemA(SESSIONA *ss, int nIndex);
void DeleteSessionItemW(SESSIONW *ss, int nIndex);
void FreeSessionItemsA(SESSIONA *ss);
void FreeSessionItemsW(SESSIONW *ss);

void FillSessionsListA(HSTACK *hStack, HWND hWnd);
void FillSessionsListW(HSTACK *hStack, HWND hWnd);
void FillItemsListA(HSTACK *hStack, HWND hWnd);
void FillItemsListW(HSTACK *hStack, HWND hWnd);
void FillItemsListCurrentA(HWND hWnd);
void FillItemsListCurrentW(HWND hWnd);

int UpdateListBoxHScrollA(HWND hWnd);
int UpdateListBoxHScrollW(HWND hWnd);
int UpdateComboBoxDropWidthA(HWND hWnd);
int UpdateComboBoxDropWidthW(HWND hWnd);
int MoveListBoxItemA(HWND hWnd, int nOldIndex, int nNewIndex);
int MoveListBoxItemW(HWND hWnd, int nOldIndex, int nNewIndex);
int MoveComboBoxItemA(HWND hWnd, int nOldIndex, int nNewIndex);
int MoveComboBoxItemW(HWND hWnd, int nOldIndex, int nNewIndex);
int GetListBoxSelItems(HWND hWnd, int **lpSelItems);
void FreeListBoxSelItems(int **lpSelItems);
BOOL IsPathFullA(char *pPath);
BOOL IsPathFullW(wchar_t *wpPath);
int GetBaseNameA(char *pFile, char *szBaseName, int nBaseNameMaxLen);
int GetBaseNameW(wchar_t *wpFile, wchar_t *wszBaseName, int nBaseNameMaxLen);
BOOL FileExistsA(char *pFile);
BOOL FileExistsW(wchar_t *wpFile);
BOOL GetWindowPos(HWND hWnd, HWND hWndOwner, RECT *rc);
BOOL DialogResizeMessages(DIALOGRESIZE *drs, RECT *rcInit, RECT *rcCurrent, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
void ReadOptionsA();
void ReadOptionsW();
void SaveOptionsA(DWORD dwFlags);
void SaveOptionsW(DWORD dwFlags);
char* GetLangStringA(LANGID wLangID, int nStringID);
wchar_t* GetLangStringW(LANGID wLangID, int nStringID);
BOOL IsExtCallParamValid(LPARAM lParam, int nIndex);
int GetExtCallParam(LPARAM lParam, int nIndex);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();

//Global variables
char szPluginName[MAX_PATH];
wchar_t wszPluginName[MAX_PATH];
HINSTANCE hInstanceDLL;
HWND hMainWnd;
HWND hMdiClient;
HWND hTab;
HMENU hMainMenu;
HICON hMainIcon;
BOOL bOldWindows;
LANGID wLangSystem;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
HANDLE hThread=NULL;
DWORD dwThreadId;
HWND hWndMainDlg=NULL;
RECT rcMainInitDialog={0};
RECT rcMainCurrentDialog={0};
BOOL bMainDialogRectSave=FALSE;
char szBuf[BUFFER_SIZE];
wchar_t wszBuf[BUFFER_SIZE];
char szSessionsDir[MAX_PATH];
wchar_t wszSessionsDir[MAX_PATH];
char szOpenOnStart[MAX_PATH]="OnExit";
wchar_t wszOpenOnStart[MAX_PATH]=L"OnExit";
char szSaveOnExit[MAX_PATH]="OnExit";
wchar_t wszSaveOnExit[MAX_PATH]=L"OnExit";
char szDlgInputBox[MAX_PATH];
wchar_t wszDlgInputBox[MAX_PATH];
HSTACK hSessionStack={0};
SESSIONA *lpVirtualSessionA=NULL;
SESSIONW *lpVirtualSessionW=NULL;
SESSIONA *lpRealSessionA=NULL;
SESSIONW *lpRealSessionW=NULL;
int nCurrentSessionIndex=CB_ERR;
BOOL bOpenOnStart=TRUE;
BOOL bSaveOnExit=TRUE;
BOOL bModalDialog=TRUE;
HMENU hMenuList=NULL;
WNDPROC OldListBoxProc;
WNDPROCDATA *NewMainProcData=NULL;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName="Sessions";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=TRUE;
    return;
  }

  if (!pd->bMDI)
  {
    if (!pd->bOnStart)
    {
      if (pd->bOldWindows)
        MessageBoxA(pd->hMainWnd, GetLangStringA(PRIMARYLANGID(pd->wLangSystem), STRID_SDI_ISNTSUPPORTED), "Sessions plugin", MB_OK|MB_ICONEXCLAMATION);
      else
        MessageBoxW(pd->hMainWnd, GetLangStringW(PRIMARYLANGID(pd->wLangSystem), STRID_SDI_ISNTSUPPORTED), L"Sessions plugin", MB_OK|MB_ICONEXCLAMATION);
    }
    return;
  }
  if (!bInitCommon) InitCommon(pd);

  if (pd->lParam)
  {
    if (IsExtCallParamValid(pd->lParam, 1))
    {
      if (bOldWindows)
      {
        if (GetExtCallParam(pd->lParam, 1) == DLLA_SESSIONS_OPEN)
        {
          char *pSession=NULL;

          if (IsExtCallParamValid(pd->lParam, 2))
            pSession=(char *)GetExtCallParam(pd->lParam, 2);

          if (pSession)
            OpenSessionFileA(pSession);
        }
      }
      else
      {
        if (GetExtCallParam(pd->lParam, 1) == DLLA_SESSIONS_OPEN)
        {
          wchar_t *wpSession=NULL;

          if (IsExtCallParamValid(pd->lParam, 2))
            wpSession=(wchar_t *)GetExtCallParam(pd->lParam, 2);

          if (wpSession)
            OpenSessionFileW(wpSession);
        }
      }
    }

    //Stay in memory, and don't change active status
    pd->nUnload=UD_NONUNLOAD_UNCHANGE;
  }
  else
  {
    if (!bInitMain) InitMain();

    if (!pd->bOnStart)
    {
      if (!hThread)
      {
        if (bOldWindows)
          hThread=CreateThread(NULL, 0, ThreadProcA, NULL, 0, &dwThreadId);
        else
          hThread=CreateThread(NULL, 0, ThreadProcW, NULL, 0, &dwThreadId);
      }
    }

    //Stay in memory, and show as active
    pd->nUnload=UD_NONUNLOAD_ACTIVE;
  }
}

DWORD WINAPI ThreadProcA(LPVOID lpParameter)
{
  MSG msg;
  BOOL bMsgStatus;

  if (hWndMainDlg=CreateDialogA(hInstanceDLL, MAKEINTRESOURCEA(IDD_SESSIONS), hMainWnd, (DLGPROC)MainDlgProcA))
  {
    if (bModalDialog)
      EnableWindow(hMainWnd, FALSE);
  }

  if (hWndMainDlg)
  {
    while ((bMsgStatus=GetMessage(&msg, NULL, 0, 0)) && bMsgStatus != -1)
    {
      if (hWndMainDlg && !IsDialogMessage(hWndMainDlg, &msg))
      {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
      }
    }
  }

  return 0;
}

DWORD WINAPI ThreadProcW(LPVOID lpParameter)
{
  MSG msg;
  BOOL bMsgStatus;

  if (hWndMainDlg=CreateDialogW(hInstanceDLL, MAKEINTRESOURCEW(IDD_SESSIONS), hMainWnd, (DLGPROC)MainDlgProcW))
  {
    if (bModalDialog)
      EnableWindow(hMainWnd, FALSE);
  }

  if (hWndMainDlg)
  {
    while ((bMsgStatus=GetMessageW(&msg, NULL, 0, 0)) && bMsgStatus != -1)
    {
      if (hWndMainDlg && !IsDialogMessageW(hWndMainDlg, &msg))
      {
        TranslateMessage(&msg);
        DispatchMessageW(&msg);
      }
    }
  }

  return 0;
}

LRESULT CALLBACK MainDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndSessionList;
  static HWND hWndItemsList;
  static HWND hWndGroupSession;
  static HWND hWndOpenButton;
  static HWND hWndCloseButton;
  static HWND hWndSaveButton;
  static HWND hWndCopyButton;
  static HWND hWndRenameButton;
  static HWND hWndDeleteButton;
  static HWND hWndSettingsButton;
  static HWND hWndOKButton;
  static SESSIONA ssCurrentTmp={0};
  static DIALOGRESIZE drs[]={{&hWndSessionList,    DRS_SIZE|DRS_X},
                             {&hWndItemsList,      DRS_SIZE|DRS_X|DRS_Y},
                             {&hWndGroupSession,   DRS_MOVE|DRS_X},
                             {&hWndOpenButton,     DRS_MOVE|DRS_X},
                             {&hWndCloseButton,    DRS_MOVE|DRS_X},
                             {&hWndSaveButton,     DRS_MOVE|DRS_X},
                             {&hWndCopyButton,     DRS_MOVE|DRS_X},
                             {&hWndRenameButton,   DRS_MOVE|DRS_X},
                             {&hWndDeleteButton,   DRS_MOVE|DRS_X},
                             {&hWndSettingsButton, DRS_MOVE|DRS_X},
                             {&hWndOKButton,       DRS_MOVE|DRS_X|DRS_Y},
                             {0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndSessionList=GetDlgItem(hDlg, IDC_SESSIONS_LIST);
    hWndItemsList=GetDlgItem(hDlg, IDC_ITEMS_LIST);
    hWndGroupSession=GetDlgItem(hDlg, IDC_GROUP_SESSION);
    hWndOpenButton=GetDlgItem(hDlg, IDC_SESSION_OPEN);
    hWndCloseButton=GetDlgItem(hDlg, IDC_SESSION_CLOSE);
    hWndSaveButton=GetDlgItem(hDlg, IDC_SESSION_SAVE);
    hWndCopyButton=GetDlgItem(hDlg, IDC_SESSION_COPY);
    hWndRenameButton=GetDlgItem(hDlg, IDC_SESSION_RENAME);
    hWndDeleteButton=GetDlgItem(hDlg, IDC_SESSION_DELETE);
    hWndSettingsButton=GetDlgItem(hDlg, IDC_SETTINGS);
    hWndOKButton=GetDlgItem(hDlg, IDOK);

    SetDlgItemTextA(hDlg, IDC_GROUP_SESSION, GetLangStringA(wLangSystem, STRID_SESSION));
    SetDlgItemTextA(hDlg, IDC_SESSION_OPEN, GetLangStringA(wLangSystem, STRID_OPEN));
    SetDlgItemTextA(hDlg, IDC_SESSION_CLOSE, GetLangStringA(wLangSystem, STRID_CLOSE));
    SetDlgItemTextA(hDlg, IDC_SESSION_SAVE, GetLangStringA(wLangSystem, STRID_SAVE));
    SetDlgItemTextA(hDlg, IDC_SESSION_COPY, GetLangStringA(wLangSystem, STRID_COPY));
    SetDlgItemTextA(hDlg, IDC_SESSION_RENAME, GetLangStringA(wLangSystem, STRID_RENAME));
    SetDlgItemTextA(hDlg, IDC_SESSION_DELETE, GetLangStringA(wLangSystem, STRID_DELETE));
    SetDlgItemTextA(hDlg, IDC_SETTINGS, GetLangStringA(wLangSystem, STRID_SETTINGS));
    SetDlgItemTextA(hDlg, IDOK, GetLangStringA(wLangSystem, STRID_OK));

    FillSessionsA(&hSessionStack);
    FillSessionsListA(&hSessionStack, hWndSessionList);
    SendMessageA(hWndSessionList, CB_INSERTSTRING, 0, (LPARAM)GetLangStringA(wLangSystem, STRID_CURRENTSESSION));
    SendMessage(hWndSessionList, CB_SETCURSEL, 0, 0);
    PostMessage(hDlg, WM_COMMAND, MAKELONG(IDC_SESSIONS_LIST, CBN_SELCHANGE), (LPARAM)hWndSessionList);

    //Popup menu
    if (hMenuList=CreatePopupMenu())
    {
      AppendMenuA(hMenuList, MF_STRING, IDC_ITEM_OPEN, GetLangStringA(wLangSystem, STRID_MENU_OPEN));
      AppendMenuA(hMenuList, MF_SEPARATOR, (UINT)-1, NULL);
      AppendMenuA(hMenuList, MF_STRING, IDC_ITEM_ADD, GetLangStringA(wLangSystem, STRID_MENU_ADD));
      AppendMenuA(hMenuList, MF_STRING, IDC_ITEM_MOVEDOWN, GetLangStringA(wLangSystem, STRID_MENU_MOVEDOWN));
      AppendMenuA(hMenuList, MF_STRING, IDC_ITEM_MOVEUP, GetLangStringA(wLangSystem, STRID_MENU_MOVEUP));
      AppendMenuA(hMenuList, MF_SEPARATOR, (UINT)-1, NULL);
      AppendMenuA(hMenuList, MF_STRING, IDC_ITEM_DELETE, GetLangStringA(wLangSystem, STRID_MENU_DELETE));
      AppendMenuA(hMenuList, MF_STRING, IDC_ITEM_DELETEOLD, GetLangStringA(wLangSystem, STRID_MENU_DELETEOLD));
    }

    //SubClass listbox
    OldListBoxProc=(WNDPROC)GetWindowLongA(hWndItemsList, GWL_WNDPROC);
    SetWindowLongA(hWndItemsList, GWL_WNDPROC, (LONG)NewListBoxProcA);
  }
  else if (uMsg == WM_CONTEXTMENU)
  {
    if ((HWND)wParam == hWndItemsList)
    {
      POINT pt;
      int nSelCount;

      if (lParam == -1)
      {
        pt.x=0;
        pt.y=0;
        ClientToScreen(hWndItemsList, &pt);
      }
      else
      {
        GetCursorPos(&pt);
      }

      nSelCount=SendMessage(hWndItemsList, LB_GETSELCOUNT, 0, 0);

      if (!lpVirtualSessionA)
      {
        ModifyMenuA(hMenuList, IDC_ITEM_OPEN, MF_BYCOMMAND|MF_STRING, IDC_ITEM_OPEN, GetLangStringA(wLangSystem, STRID_MENU_ACTIVATE));
        EnableMenuItem(hMenuList, IDC_ITEM_OPEN, (nSelCount == 1)?MF_ENABLED:MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_ADD, MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_MOVEDOWN, MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_MOVEUP, MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_DELETE, MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_DELETEOLD, MF_GRAYED);
      }
      else
      {
        ModifyMenuA(hMenuList, IDC_ITEM_OPEN, MF_BYCOMMAND|MF_STRING, IDC_ITEM_OPEN, GetLangStringA(wLangSystem, STRID_MENU_OPEN));
        EnableMenuItem(hMenuList, IDC_ITEM_OPEN, nSelCount?MF_ENABLED:MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_ADD, MF_ENABLED);
        EnableMenuItem(hMenuList, IDC_ITEM_MOVEDOWN, nSelCount?MF_ENABLED:MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_MOVEUP, nSelCount?MF_ENABLED:MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_DELETE, nSelCount?MF_ENABLED:MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_DELETEOLD, MF_ENABLED);
      }
      TrackPopupMenu(hMenuList, TPM_LEFTBUTTON|TPM_RIGHTBUTTON, pt.x, pt.y, 0, hDlg, NULL);
    }
  }
  else if (uMsg == AKDLL_UPDATESAVEBUTTON)
  {
    BOOL bEnable=FALSE;

    if (!lpVirtualSessionA || lpVirtualSessionA->bModified)
      bEnable=TRUE;
    EnableWindow(hWndSaveButton, bEnable);
    EnableWindow(hWndCopyButton, !bEnable);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_SESSIONS_LIST)
    {
      if (HIWORD(wParam) == CBN_SELCHANGE)
      {
        if (SaveSessionPromptA() == IDCANCEL)
        {
          SendMessage(hWndSessionList, CB_SETCURSEL, nCurrentSessionIndex, 0);
          return 0;
        }
        if (lpVirtualSessionA)
          FreeSessionItemsA(lpVirtualSessionA);
        lpVirtualSessionA=NULL;
        lpRealSessionA=NULL;

        if (!(nCurrentSessionIndex=SendMessage(hWndSessionList, CB_GETCURSEL, 0, 0)))
        {
          SendMessage(hWndItemsList, LB_RESETCONTENT, 0, 0);
          FillItemsListCurrentA(hWndItemsList);
          UpdateListBoxHScrollA(hWndItemsList);

          if (SendMessage(hWndItemsList, LB_GETCOUNT, 0, 0) > 0)
            EnableWindow(hWndCloseButton, TRUE);
          else
            EnableWindow(hWndCloseButton, FALSE);
          EnableWindow(hWndOpenButton, FALSE);
          EnableWindow(hWndCopyButton, FALSE);
          EnableWindow(hWndRenameButton, FALSE);
          EnableWindow(hWndDeleteButton, FALSE);
        }
        else
        {
          SendMessage(hWndItemsList, LB_RESETCONTENT, 0, 0);
          SendMessageA(hWndSessionList, CB_GETLBTEXT, nCurrentSessionIndex, (LPARAM)szBuf);

          if (lpRealSessionA=GetSessionA(&hSessionStack, szBuf))
          {
            FillItemsListA(&lpRealSessionA->hItemsStack, hWndItemsList);
            CopySessionA(lpRealSessionA, &ssCurrentTmp);
            lpVirtualSessionA=&ssCurrentTmp;
            UpdateListBoxHScrollA(hWndItemsList);
          }

          if (SendMessage(hWndItemsList, LB_GETCOUNT, 0, 0) > 0)
          {
            EnableWindow(hWndOpenButton, TRUE);
            EnableWindow(hWndCloseButton, TRUE);
          }
          else
          {
            EnableWindow(hWndOpenButton, FALSE);
            EnableWindow(hWndCloseButton, FALSE);
          }
          EnableWindow(hWndCopyButton, TRUE);
          EnableWindow(hWndRenameButton, TRUE);
          EnableWindow(hWndDeleteButton, TRUE);
        }
        PostMessage(hDlg, AKDLL_UPDATESAVEBUTTON, 0, 0);
      }
    }
    else if (LOWORD(wParam) == IDC_ITEMS_LIST)
    {
      if (HIWORD(wParam) == LBN_DBLCLK)
      {
        PostMessage(hDlg, WM_COMMAND, IDC_ITEM_OPEN, 0);
      }
    }
    else if (LOWORD(wParam) == IDC_SESSION_OPEN)
    {
      OpenSessionA(&lpVirtualSessionA->hItemsStack);
    }
    else if (LOWORD(wParam) == IDC_SESSION_CLOSE)
    {
      if (!lpVirtualSessionA)
      {
        if (bModalDialog) EnableWindow(hMainWnd, TRUE);
        ShowWindow(hDlg, FALSE);
        SendMessage(hMainWnd, WM_COMMAND, IDM_FILE_CLOSEALL, 0);
        SendMessage(hDlg, WM_COMMAND, MAKELONG(IDC_SESSIONS_LIST, CBN_SELCHANGE), (LPARAM)hWndSessionList);
        ShowWindow(hDlg, TRUE);
        if (bModalDialog) EnableWindow(hMainWnd, FALSE);
      }
      else CloseSessionA(&lpVirtualSessionA->hItemsStack);
    }
    else if (LOWORD(wParam) == IDC_SESSION_SAVE ||
             LOWORD(wParam) == IDC_SESSION_COPY)
    {
      SESSIONA *lpNewElement;
      SESSIONA *lpOldElement;
      char szNewSession[MAX_PATH];
      int nIndex;
      BOOL bSave=TRUE;

      if (lpVirtualSessionA && LOWORD(wParam) == IDC_SESSION_SAVE)
      {
        FreeSessionItemsA(lpRealSessionA);
        lpVirtualSessionA->bModified=FALSE;
        CopySessionA(lpVirtualSessionA, lpRealSessionA);
        SaveSessionFileA(lpRealSessionA);
        PostMessage(hDlg, AKDLL_UPDATESAVEBUTTON, 0, 0);
        return 0;
      }

      if (!DialogBoxParamA(hInstanceDLL, MAKEINTRESOURCEA(IDD_INPUTBOX), hDlg, (DLGPROC)InputBoxDlgProcA, LOWORD(wParam)))
        return 0;
      lstrcpynA(szNewSession, szDlgInputBox, MAX_PATH);

      if (lpOldElement=GetSessionA(&hSessionStack, szNewSession))
      {
        wsprintfA(szBuf, GetLangStringA(wLangSystem, STRID_ALREADY_EXIST), szNewSession);
        if (MessageBoxA(hDlg, szBuf, "Sessions plugin", MB_YESNO|MB_ICONQUESTION|MB_DEFBUTTON2) == IDNO)
          bSave=FALSE;
      }

      if (bSave)
      {
        if (!lpVirtualSessionA)
        {
          lpNewElement=AddCurrentSessionA(&hSessionStack, szNewSession);
        }
        else
        {
          if (lpNewElement=AddEmptySessionA(&hSessionStack, szNewSession))
          {
            lstrcpynA(lpVirtualSessionA->szSessionName, szNewSession, MAX_PATH);
            lpVirtualSessionA->bModified=FALSE;
            CopySessionA(lpVirtualSessionA, lpNewElement);
          }
        }
        SaveSessionFileA(lpNewElement);

        if (!lpOldElement)
        {
          if ((nIndex=SendMessageA(hWndSessionList, CB_ADDSTRING, 0, (LPARAM)szNewSession)) != CB_ERR)
          {
            if (!nIndex)
              MoveComboBoxItemA(hWndSessionList, nIndex, ++nIndex);
          }
        }
        else
        {
          DeleteSessionA(&hSessionStack, lpOldElement);
          nIndex=SendMessageA(hWndSessionList, CB_FINDSTRINGEXACT, (WPARAM)-1, (LPARAM)szNewSession);
        }
        SetFocus(hWndSessionList);
        SendMessage(hWndSessionList, CB_SETCURSEL, nIndex, 0);
        PostMessage(hDlg, WM_COMMAND, MAKELONG(IDC_SESSIONS_LIST, CBN_SELCHANGE), (LPARAM)hWndSessionList);
      }
    }
    else if (LOWORD(wParam) == IDC_SESSION_RENAME)
    {
      if (DialogBoxParamA(hInstanceDLL, MAKEINTRESOURCEA(IDD_INPUTBOX), hDlg, (DLGPROC)InputBoxDlgProcA, IDC_SESSION_RENAME))
      {
        if (lpRealSessionA && lpVirtualSessionA)
        {
          if (RenameSessionFileA(lpRealSessionA->szSessionName, szDlgInputBox))
          {
            lstrcpynA(lpVirtualSessionA->szSessionName, szDlgInputBox, MAX_PATH);
            lstrcpynA(lpRealSessionA->szSessionName, szDlgInputBox, MAX_PATH);

            SendMessage(hWndSessionList, CB_DELETESTRING, nCurrentSessionIndex, 0);
            SendMessageA(hWndSessionList, CB_INSERTSTRING, nCurrentSessionIndex, (LPARAM)lpVirtualSessionA->szSessionName);
            SetFocus(hWndSessionList);
            SendMessage(hWndSessionList, CB_SETCURSEL, nCurrentSessionIndex, 0);
          }
          else MessageBoxA(hDlg, GetLangStringA(wLangSystem, STRID_RENAME_ERROR), "Sessions plugin", MB_OK|MB_ICONERROR);
        }
      }
    }
    else if (LOWORD(wParam) == IDC_SESSION_DELETE)
    {
      if (lpVirtualSessionA)
      {
        if (MessageBoxA(hDlg, GetLangStringA(wLangSystem, STRID_CONFIRM_DELETE), "Sessions plugin", MB_YESNO|MB_ICONQUESTION) == IDYES)
        {
          if (DeleteSessionFileA(lpVirtualSessionA->szSessionName))
          {
            //Delete session from stack
            DeleteSessionA(&hSessionStack, lpRealSessionA);
            SendMessage(hWndSessionList, CB_DELETESTRING, nCurrentSessionIndex, 0);
            lpRealSessionA=NULL;

            //Delete temp session
            FreeSessionItemsA(lpVirtualSessionA);
            lpVirtualSessionA=NULL;

            SetFocus(hWndSessionList);
            SendMessage(hWndSessionList, CB_SETCURSEL, 0, 0);
            PostMessage(hDlg, WM_COMMAND, MAKELONG(IDC_SESSIONS_LIST, CBN_SELCHANGE), (LPARAM)hWndSessionList);
          }
        }
      }
    }
    else if (LOWORD(wParam) == IDC_ITEM_OPEN)
    {
      int *lpSelItems;
      int nSelCount;

      if (nSelCount=GetListBoxSelItems(hWndItemsList, &lpSelItems))
      {
        if (!lpVirtualSessionA)
        {
          TCITEMA tcItem;
          EDITINFO ei;
          int nNonEmpty=0;
          int nIndex=0;

          while (1)
          {
            tcItem.mask=TCIF_PARAM;
            if (!SendMessage(hTab, TCM_GETITEMA, nIndex++, (LPARAM)&tcItem)) break;

            if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)GetDlgItem((HWND)tcItem.lParam, ID_EDIT), (LPARAM)&ei))
            {
              if (*(char *)ei.pFile)
              {
                if (lpSelItems[0] == nNonEmpty)
                {
                  SendMessage(hMdiClient, WM_MDIACTIVATE, (WPARAM)tcItem.lParam, 0);
                  break;
                }
                ++nNonEmpty;
              }
            }
          }
        }
        else
        {
          SESSIONITEMA *lpElement;
          int i;

          for (i=0; i < nSelCount; ++i)
          {
            if (lpElement=GetSessionItemA(lpVirtualSessionA, lpSelItems[i] + 1))
            {
              OpenSessionItemA(lpElement);
            }
          }
        }
        FreeListBoxSelItems(&lpSelItems);
      }
    }
    else if (LOWORD(wParam) == IDC_ITEM_ADD)
    {
      if (lpVirtualSessionA)
      {
        OPENFILENAMEA ofn={0};

        szBuf[0]='\0';
        ofn.lStructSize  =sizeof(OPENFILENAMEA);
        ofn.hwndOwner    =hDlg;
        ofn.hInstance    =hInstanceDLL;
        ofn.lpstrFile    =szBuf;
        ofn.lpstrFilter  =GetLangStringA(wLangSystem, STRID_FILTER);
        ofn.nFilterIndex =2;
        ofn.nMaxFile     =BUFFER_SIZE;
        ofn.Flags        =OFN_ALLOWMULTISELECT|OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_EXPLORER|OFN_ENABLESIZING;

        if (GetOpenFileNameA(&ofn))
        {
          char szFile[MAX_PATH];
          char *pFile=szBuf + lstrlenA(szBuf) + 1;
          SESSIONITEMA *lpItemElement;
          int nIndex;

          SendMessage(hWndItemsList, LB_SETSEL, FALSE, -1);

          if (*pFile)
          {
            if (*(pFile - 2) == '\\') *(pFile - 2)='\0';

            do
            {
              if (IsPathFullA(pFile))
                lstrcpyA(szFile, pFile);  //.lnk target
              else
                wsprintfA(szFile, "%s\\%s", szBuf, pFile);

              if (lpItemElement=AddSessionItemA(lpVirtualSessionA))
              {
                lstrcpynA(lpItemElement->szItemFile, szFile, MAX_PATH);
                nIndex=SendMessageA(hWndItemsList, LB_ADDSTRING, 0, (LPARAM)szFile);
                SendMessage(hWndItemsList, LB_SETSEL, TRUE, nIndex);
                lpVirtualSessionA->bModified=TRUE;
              }
            }
            while (*(pFile+=lstrlenA(pFile) + 1));
          }
          else
          {
            if (lpItemElement=AddSessionItemA(lpVirtualSessionA))
            {
              lstrcpynA(lpItemElement->szItemFile, szBuf, MAX_PATH);
              nIndex=SendMessageA(hWndItemsList, LB_ADDSTRING, 0, (LPARAM)szBuf);
              SendMessage(hWndItemsList, LB_SETSEL, TRUE, nIndex);
              lpVirtualSessionA->bModified=TRUE;
            }
          }
          PostMessage(hDlg, AKDLL_UPDATESAVEBUTTON, 0, 0);
        }
      }
    }
    else if (LOWORD(wParam) == IDC_ITEM_MOVEDOWN)
    {
      if (lpVirtualSessionA)
      {
        int *lpSelItems;
        int nSelCount;
        int nMaxIndex;
        int nOldIndex=-1;
        int nNewIndex=-1;
        int i;

        nMaxIndex=SendMessage(hWndItemsList, LB_GETCOUNT, 0, 0) - 1;

        if (nSelCount=GetListBoxSelItems(hWndItemsList, &lpSelItems))
        {
          for (i=--nSelCount; i >= 0; --i)
          {
            if (lpSelItems[i] < nMaxIndex)
            {
              if (nNewIndex == -1 && i < nSelCount)
              {
                if (lpSelItems[i] + 1 >= lpSelItems[i + 1])
                  continue;
              }
              nOldIndex=lpSelItems[i];
              nNewIndex=lpSelItems[i] + 1;
              MoveListBoxItemA(hWndItemsList, nOldIndex, nNewIndex);
              MoveSessionItemA(lpVirtualSessionA, nOldIndex + 1, nNewIndex + 1);
              SendMessage(hWndItemsList, LB_SETSEL, TRUE, nNewIndex);
              lpVirtualSessionA->bModified=TRUE;
            }
          }
          FreeListBoxSelItems(&lpSelItems);
        }
        PostMessage(hDlg, AKDLL_UPDATESAVEBUTTON, 0, 0);
      }
    }
    else if (LOWORD(wParam) == IDC_ITEM_MOVEUP)
    {
      if (lpVirtualSessionA)
      {
        int *lpSelItems;
        int nSelCount;
        int nMinIndex;
        int nOldIndex=-1;
        int nNewIndex=-1;
        int i;

        nMinIndex=0;

        if (nSelCount=GetListBoxSelItems(hWndItemsList, &lpSelItems))
        {
          for (i=0; i < nSelCount; ++i)
          {
            if (lpSelItems[i] > nMinIndex)
            {
              if (nNewIndex == -1 && i > 0)
              {
                if (lpSelItems[i] - 1 <= lpSelItems[i - 1])
                  continue;
              }
              nOldIndex=lpSelItems[i];
              nNewIndex=lpSelItems[i] - 1;
              MoveListBoxItemA(hWndItemsList, nOldIndex, nNewIndex);
              MoveSessionItemA(lpVirtualSessionA, nOldIndex + 1, nNewIndex + 1);
              SendMessage(hWndItemsList, LB_SETSEL, TRUE, nNewIndex);
              lpVirtualSessionA->bModified=TRUE;
            }
          }
          FreeListBoxSelItems(&lpSelItems);
        }
        PostMessage(hDlg, AKDLL_UPDATESAVEBUTTON, 0, 0);
      }
    }
    else if (LOWORD(wParam) == IDC_ITEM_DELETE)
    {
      if (lpVirtualSessionA)
      {
        int *lpSelItems;
        int nSelCount;
        int i;

        if (nSelCount=GetListBoxSelItems(hWndItemsList, &lpSelItems))
        {
          for (i=--nSelCount; i >= 0; --i)
          {
            SendMessage(hWndItemsList, LB_DELETESTRING, lpSelItems[i], 0);
            DeleteSessionItemA(lpVirtualSessionA, lpSelItems[i] + 1);
            lpVirtualSessionA->bModified=TRUE;
          }
          FreeListBoxSelItems(&lpSelItems);
        }
        PostMessage(hDlg, AKDLL_UPDATESAVEBUTTON, 0, 0);
      }
    }
    else if (LOWORD(wParam) == IDC_ITEM_DELETEOLD)
    {
      if (lpVirtualSessionA)
      {
        SESSIONITEMA *lpElement;
        int nMaxIndex;
        int i;

        nMaxIndex=SendMessage(hWndItemsList, LB_GETCOUNT, 0, 0) - 1;

        for (i=nMaxIndex; i >= 0; --i)
        {
          if (lpElement=GetSessionItemA(lpVirtualSessionA, i + 1))
          {
            if (!FileExistsA(lpElement->szItemFile))
            {
              SendMessage(hWndItemsList, LB_DELETESTRING, i, 0);
              DeleteSessionItemA(lpVirtualSessionA, i + 1);
              lpVirtualSessionA->bModified=TRUE;
            }
          }
          else break;
        }
        PostMessage(hDlg, AKDLL_UPDATESAVEBUTTON, 0, 0);
      }
    }
    else if (LOWORD(wParam) == IDC_SETTINGS)
    {
      DialogBoxA(hInstanceDLL, MAKEINTRESOURCEA(IDD_SETTINGS), hDlg, (DLGPROC)SettingsDlgProcA);
    }
    else if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL)
    {
      if (SaveSessionPromptA() == IDCANCEL)
        return 0;
      if (lpVirtualSessionA)
        FreeSessionItemsA(lpVirtualSessionA);
      lpVirtualSessionA=NULL;
      lpRealSessionA=NULL;

      FreeSessionsA(&hSessionStack);
      DestroyMenu(hMenuList);

      if (!IsWindowEnabled(hMainWnd)) EnableWindow(hMainWnd, TRUE);
      DestroyWindow(hWndMainDlg);
      hWndMainDlg=NULL;
      CloseHandle(hThread);
      hThread=NULL;
      ExitThread(0);
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }

  if (DialogResizeMessages(&drs[0], &rcMainInitDialog, &rcMainCurrentDialog, hDlg, uMsg, wParam, lParam))
    bMainDialogRectSave=TRUE;

  return FALSE;
}

LRESULT CALLBACK MainDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndSessionList;
  static HWND hWndItemsList;
  static HWND hWndGroupSession;
  static HWND hWndOpenButton;
  static HWND hWndCloseButton;
  static HWND hWndSaveButton;
  static HWND hWndCopyButton;
  static HWND hWndRenameButton;
  static HWND hWndDeleteButton;
  static HWND hWndSettingsButton;
  static HWND hWndOKButton;
  static SESSIONW ssCurrentTmp={0};
  static DIALOGRESIZE drs[]={{&hWndSessionList,    DRS_SIZE|DRS_X},
                             {&hWndItemsList,      DRS_SIZE|DRS_X|DRS_Y},
                             {&hWndGroupSession,   DRS_MOVE|DRS_X},
                             {&hWndOpenButton,     DRS_MOVE|DRS_X},
                             {&hWndCloseButton,    DRS_MOVE|DRS_X},
                             {&hWndSaveButton,     DRS_MOVE|DRS_X},
                             {&hWndCopyButton,     DRS_MOVE|DRS_X},
                             {&hWndRenameButton,   DRS_MOVE|DRS_X},
                             {&hWndDeleteButton,   DRS_MOVE|DRS_X},
                             {&hWndSettingsButton, DRS_MOVE|DRS_X},
                             {&hWndOKButton,       DRS_MOVE|DRS_X|DRS_Y},
                             {0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndSessionList=GetDlgItem(hDlg, IDC_SESSIONS_LIST);
    hWndItemsList=GetDlgItem(hDlg, IDC_ITEMS_LIST);
    hWndGroupSession=GetDlgItem(hDlg, IDC_GROUP_SESSION);
    hWndOpenButton=GetDlgItem(hDlg, IDC_SESSION_OPEN);
    hWndCloseButton=GetDlgItem(hDlg, IDC_SESSION_CLOSE);
    hWndSaveButton=GetDlgItem(hDlg, IDC_SESSION_SAVE);
    hWndCopyButton=GetDlgItem(hDlg, IDC_SESSION_COPY);
    hWndRenameButton=GetDlgItem(hDlg, IDC_SESSION_RENAME);
    hWndDeleteButton=GetDlgItem(hDlg, IDC_SESSION_DELETE);
    hWndSettingsButton=GetDlgItem(hDlg, IDC_SETTINGS);
    hWndOKButton=GetDlgItem(hDlg, IDOK);

    SetDlgItemTextW(hDlg, IDC_GROUP_SESSION, GetLangStringW(wLangSystem, STRID_SESSION));
    SetDlgItemTextW(hDlg, IDC_SESSION_OPEN, GetLangStringW(wLangSystem, STRID_OPEN));
    SetDlgItemTextW(hDlg, IDC_SESSION_CLOSE, GetLangStringW(wLangSystem, STRID_CLOSE));
    SetDlgItemTextW(hDlg, IDC_SESSION_SAVE, GetLangStringW(wLangSystem, STRID_SAVE));
    SetDlgItemTextW(hDlg, IDC_SESSION_COPY, GetLangStringW(wLangSystem, STRID_COPY));
    SetDlgItemTextW(hDlg, IDC_SESSION_RENAME, GetLangStringW(wLangSystem, STRID_RENAME));
    SetDlgItemTextW(hDlg, IDC_SESSION_DELETE, GetLangStringW(wLangSystem, STRID_DELETE));
    SetDlgItemTextW(hDlg, IDC_SETTINGS, GetLangStringW(wLangSystem, STRID_SETTINGS));
    SetDlgItemTextW(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));

    FillSessionsW(&hSessionStack);
    FillSessionsListW(&hSessionStack, hWndSessionList);
    SendMessageW(hWndSessionList, CB_INSERTSTRING, 0, (LPARAM)GetLangStringW(wLangSystem, STRID_CURRENTSESSION));
    SendMessage(hWndSessionList, CB_SETCURSEL, 0, 0);
    PostMessage(hDlg, WM_COMMAND, MAKELONG(IDC_SESSIONS_LIST, CBN_SELCHANGE), (LPARAM)hWndSessionList);

    //Popup menu
    if (hMenuList=CreatePopupMenu())
    {
      AppendMenuW(hMenuList, MF_STRING, IDC_ITEM_OPEN, GetLangStringW(wLangSystem, STRID_MENU_OPEN));
      AppendMenuW(hMenuList, MF_SEPARATOR, (UINT)-1, NULL);
      AppendMenuW(hMenuList, MF_STRING, IDC_ITEM_ADD, GetLangStringW(wLangSystem, STRID_MENU_ADD));
      AppendMenuW(hMenuList, MF_STRING, IDC_ITEM_MOVEDOWN, GetLangStringW(wLangSystem, STRID_MENU_MOVEDOWN));
      AppendMenuW(hMenuList, MF_STRING, IDC_ITEM_MOVEUP, GetLangStringW(wLangSystem, STRID_MENU_MOVEUP));
      AppendMenuW(hMenuList, MF_SEPARATOR, (UINT)-1, NULL);
      AppendMenuW(hMenuList, MF_STRING, IDC_ITEM_DELETE, GetLangStringW(wLangSystem, STRID_MENU_DELETE));
      AppendMenuW(hMenuList, MF_STRING, IDC_ITEM_DELETEOLD, GetLangStringW(wLangSystem, STRID_MENU_DELETEOLD));
    }

    //SubClass listbox
    OldListBoxProc=(WNDPROC)GetWindowLongW(hWndItemsList, GWL_WNDPROC);
    SetWindowLongW(hWndItemsList, GWL_WNDPROC, (LONG)NewListBoxProcW);
  }
  else if (uMsg == WM_CONTEXTMENU)
  {
    if ((HWND)wParam == hWndItemsList)
    {
      POINT pt;
      int nSelCount;

      if (lParam == -1)
      {
        pt.x=0;
        pt.y=0;
        ClientToScreen(hWndItemsList, &pt);
      }
      else
      {
        GetCursorPos(&pt);
      }

      nSelCount=SendMessage(hWndItemsList, LB_GETSELCOUNT, 0, 0);

      if (!lpVirtualSessionW)
      {
        ModifyMenuW(hMenuList, IDC_ITEM_OPEN, MF_BYCOMMAND|MF_STRING, IDC_ITEM_OPEN, GetLangStringW(wLangSystem, STRID_MENU_ACTIVATE));
        EnableMenuItem(hMenuList, IDC_ITEM_OPEN, (nSelCount == 1)?MF_ENABLED:MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_ADD, MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_MOVEDOWN, MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_MOVEUP, MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_DELETE, MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_DELETEOLD, MF_GRAYED);
      }
      else
      {
        ModifyMenuW(hMenuList, IDC_ITEM_OPEN, MF_BYCOMMAND|MF_STRING, IDC_ITEM_OPEN, GetLangStringW(wLangSystem, STRID_MENU_OPEN));
        EnableMenuItem(hMenuList, IDC_ITEM_OPEN, nSelCount?MF_ENABLED:MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_ADD, MF_ENABLED);
        EnableMenuItem(hMenuList, IDC_ITEM_MOVEDOWN, nSelCount?MF_ENABLED:MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_MOVEUP, nSelCount?MF_ENABLED:MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_DELETE, nSelCount?MF_ENABLED:MF_GRAYED);
        EnableMenuItem(hMenuList, IDC_ITEM_DELETEOLD, MF_ENABLED);
      }
      TrackPopupMenu(hMenuList, TPM_LEFTBUTTON|TPM_RIGHTBUTTON, pt.x, pt.y, 0, hDlg, NULL);
    }
  }
  else if (uMsg == AKDLL_UPDATESAVEBUTTON)
  {
    BOOL bEnable=FALSE;

    if (!lpVirtualSessionW || lpVirtualSessionW->bModified)
      bEnable=TRUE;
    EnableWindow(hWndSaveButton, bEnable);
    EnableWindow(hWndCopyButton, !bEnable);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_SESSIONS_LIST)
    {
      if (HIWORD(wParam) == CBN_SELCHANGE)
      {
        if (SaveSessionPromptW() == IDCANCEL)
        {
          SendMessage(hWndSessionList, CB_SETCURSEL, nCurrentSessionIndex, 0);
          return 0;
        }
        if (lpVirtualSessionW)
          FreeSessionItemsW(lpVirtualSessionW);
        lpVirtualSessionW=NULL;
        lpRealSessionW=NULL;

        if (!(nCurrentSessionIndex=SendMessage(hWndSessionList, CB_GETCURSEL, 0, 0)))
        {
          SendMessage(hWndItemsList, LB_RESETCONTENT, 0, 0);
          FillItemsListCurrentW(hWndItemsList);
          UpdateListBoxHScrollW(hWndItemsList);

          if (SendMessage(hWndItemsList, LB_GETCOUNT, 0, 0) > 0)
            EnableWindow(hWndCloseButton, TRUE);
          else
            EnableWindow(hWndCloseButton, FALSE);
          EnableWindow(hWndOpenButton, FALSE);
          EnableWindow(hWndCopyButton, FALSE);
          EnableWindow(hWndRenameButton, FALSE);
          EnableWindow(hWndDeleteButton, FALSE);
        }
        else
        {
          SendMessage(hWndItemsList, LB_RESETCONTENT, 0, 0);
          SendMessageW(hWndSessionList, CB_GETLBTEXT, nCurrentSessionIndex, (LPARAM)wszBuf);

          if (lpRealSessionW=GetSessionW(&hSessionStack, wszBuf))
          {
            FillItemsListW(&lpRealSessionW->hItemsStack, hWndItemsList);
            CopySessionW(lpRealSessionW, &ssCurrentTmp);
            lpVirtualSessionW=&ssCurrentTmp;
            UpdateListBoxHScrollW(hWndItemsList);
          }

          if (SendMessage(hWndItemsList, LB_GETCOUNT, 0, 0) > 0)
          {
            EnableWindow(hWndOpenButton, TRUE);
            EnableWindow(hWndCloseButton, TRUE);
          }
          else
          {
            EnableWindow(hWndOpenButton, FALSE);
            EnableWindow(hWndCloseButton, FALSE);
          }
          EnableWindow(hWndCopyButton, TRUE);
          EnableWindow(hWndRenameButton, TRUE);
          EnableWindow(hWndDeleteButton, TRUE);
        }
        PostMessage(hDlg, AKDLL_UPDATESAVEBUTTON, 0, 0);
      }
    }
    else if (LOWORD(wParam) == IDC_ITEMS_LIST)
    {
      if (HIWORD(wParam) == LBN_DBLCLK)
      {
        PostMessage(hDlg, WM_COMMAND, IDC_ITEM_OPEN, 0);
      }
    }
    else if (LOWORD(wParam) == IDC_SESSION_OPEN)
    {
      OpenSessionW(&lpVirtualSessionW->hItemsStack);
    }
    else if (LOWORD(wParam) == IDC_SESSION_CLOSE)
    {
      if (!lpVirtualSessionW)
      {
        if (bModalDialog) EnableWindow(hMainWnd, TRUE);
        ShowWindow(hDlg, FALSE);
        SendMessage(hMainWnd, WM_COMMAND, IDM_FILE_CLOSEALL, 0);
        SendMessage(hDlg, WM_COMMAND, MAKELONG(IDC_SESSIONS_LIST, CBN_SELCHANGE), (LPARAM)hWndSessionList);
        ShowWindow(hDlg, TRUE);
        if (bModalDialog) EnableWindow(hMainWnd, FALSE);
      }
      else CloseSessionW(&lpVirtualSessionW->hItemsStack);
    }
    else if (LOWORD(wParam) == IDC_SESSION_SAVE ||
             LOWORD(wParam) == IDC_SESSION_COPY)
    {
      SESSIONW *lpNewElement;
      SESSIONW *lpOldElement=NULL;
      wchar_t wszNewSession[MAX_PATH];
      int nIndex;
      BOOL bSave=TRUE;

      if (lpVirtualSessionW && LOWORD(wParam) == IDC_SESSION_SAVE)
      {
        FreeSessionItemsW(lpRealSessionW);
        lpVirtualSessionW->bModified=FALSE;
        CopySessionW(lpVirtualSessionW, lpRealSessionW);
        SaveSessionFileW(lpRealSessionW);
        PostMessage(hDlg, AKDLL_UPDATESAVEBUTTON, 0, 0);
        return 0;
      }

      if (!DialogBoxParamW(hInstanceDLL, MAKEINTRESOURCEW(IDD_INPUTBOX), hDlg, (DLGPROC)InputBoxDlgProcW, LOWORD(wParam)))
        return 0;
      lstrcpynW(wszNewSession, wszDlgInputBox, MAX_PATH);

      if (lpOldElement=GetSessionW(&hSessionStack, wszNewSession))
      {
        wsprintfW(wszBuf, GetLangStringW(wLangSystem, STRID_ALREADY_EXIST), wszNewSession);
        if (MessageBoxW(hDlg, wszBuf, L"Sessions plugin", MB_YESNO|MB_ICONQUESTION|MB_DEFBUTTON2) == IDNO)
          bSave=FALSE;
      }

      if (bSave)
      {
        if (!lpVirtualSessionW)
        {
          lpNewElement=AddCurrentSessionW(&hSessionStack, wszNewSession);
        }
        else
        {
          if (lpNewElement=AddEmptySessionW(&hSessionStack, wszNewSession))
          {
            lstrcpynW(lpVirtualSessionW->wszSessionName, wszNewSession, MAX_PATH);
            lpVirtualSessionW->bModified=FALSE;
            CopySessionW(lpVirtualSessionW, lpNewElement);
          }
        }
        SaveSessionFileW(lpNewElement);

        if (!lpOldElement)
        {
          if ((nIndex=SendMessageW(hWndSessionList, CB_ADDSTRING, 0, (LPARAM)wszNewSession)) != CB_ERR)
          {
            if (!nIndex)
              MoveComboBoxItemW(hWndSessionList, nIndex, ++nIndex);
          }
        }
        else
        {
          DeleteSessionW(&hSessionStack, lpOldElement);
          nIndex=SendMessageW(hWndSessionList, CB_FINDSTRINGEXACT, (WPARAM)-1, (LPARAM)wszNewSession);
        }
        SetFocus(hWndSessionList);
        SendMessage(hWndSessionList, CB_SETCURSEL, nIndex, 0);
        PostMessage(hDlg, WM_COMMAND, MAKELONG(IDC_SESSIONS_LIST, CBN_SELCHANGE), (LPARAM)hWndSessionList);
      }
    }
    else if (LOWORD(wParam) == IDC_SESSION_RENAME)
    {
      if (DialogBoxParamW(hInstanceDLL, MAKEINTRESOURCEW(IDD_INPUTBOX), hDlg, (DLGPROC)InputBoxDlgProcW, IDC_SESSION_RENAME))
      {
        if (lpRealSessionW && lpVirtualSessionW)
        {
          if (RenameSessionFileW(lpRealSessionW->wszSessionName, wszDlgInputBox))
          {
            lstrcpynW(lpVirtualSessionW->wszSessionName, wszDlgInputBox, MAX_PATH);
            lstrcpynW(lpRealSessionW->wszSessionName, wszDlgInputBox, MAX_PATH);

            SendMessage(hWndSessionList, CB_DELETESTRING, nCurrentSessionIndex, 0);
            SendMessageW(hWndSessionList, CB_INSERTSTRING, nCurrentSessionIndex, (LPARAM)lpVirtualSessionW->wszSessionName);
            SetFocus(hWndSessionList);
            SendMessage(hWndSessionList, CB_SETCURSEL, nCurrentSessionIndex, 0);
          }
          else MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_RENAME_ERROR), L"Sessions plugin", MB_OK|MB_ICONERROR);
        }
      }
    }
    else if (LOWORD(wParam) == IDC_SESSION_DELETE)
    {
      if (lpVirtualSessionW)
      {
        if (MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_CONFIRM_DELETE), L"Sessions plugin", MB_YESNO|MB_ICONQUESTION) == IDYES)
        {
          if (DeleteSessionFileW(lpVirtualSessionW->wszSessionName))
          {
            //Delete session from stack
            DeleteSessionW(&hSessionStack, lpRealSessionW);
            SendMessage(hWndSessionList, CB_DELETESTRING, nCurrentSessionIndex, 0);
            lpRealSessionW=NULL;

            //Delete temp session
            FreeSessionItemsW(lpVirtualSessionW);
            lpVirtualSessionW=NULL;

            SetFocus(hWndSessionList);
            SendMessage(hWndSessionList, CB_SETCURSEL, 0, 0);
            PostMessage(hDlg, WM_COMMAND, MAKELONG(IDC_SESSIONS_LIST, CBN_SELCHANGE), (LPARAM)hWndSessionList);
          }
        }
      }
    }
    else if (LOWORD(wParam) == IDC_ITEM_OPEN)
    {
      int *lpSelItems;
      int nSelCount;

      if (nSelCount=GetListBoxSelItems(hWndItemsList, &lpSelItems))
      {
        if (!lpVirtualSessionW)
        {
          TCITEMW tcItem;
          EDITINFO ei;
          int nNonEmpty=0;
          int nIndex=0;

          while (1)
          {
            tcItem.mask=TCIF_PARAM;
            if (!SendMessage(hTab, TCM_GETITEMW, nIndex++, (LPARAM)&tcItem)) break;

            if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)GetDlgItem((HWND)tcItem.lParam, ID_EDIT), (LPARAM)&ei))
            {
              if (*(wchar_t *)ei.pFile)
              {
                if (lpSelItems[0] == nNonEmpty)
                {
                  SendMessage(hMdiClient, WM_MDIACTIVATE, (WPARAM)tcItem.lParam, 0);
                  break;
                }
                ++nNonEmpty;
              }
            }
          }
        }
        else
        {
          SESSIONITEMW *lpElement;
          int i;

          for (i=0; i < nSelCount; ++i)
          {
            if (lpElement=GetSessionItemW(lpVirtualSessionW, lpSelItems[i] + 1))
            {
              OpenSessionItemW(lpElement);
            }
          }
        }
        FreeListBoxSelItems(&lpSelItems);
      }
    }
    else if (LOWORD(wParam) == IDC_ITEM_ADD)
    {
      if (lpVirtualSessionW)
      {
        OPENFILENAMEW ofn={0};

        wszBuf[0]='\0';
        ofn.lStructSize  =sizeof(OPENFILENAMEW);
        ofn.hwndOwner    =hDlg;
        ofn.hInstance    =hInstanceDLL;
        ofn.lpstrFile    =wszBuf;
        ofn.lpstrFilter  =GetLangStringW(wLangSystem, STRID_FILTER);
        ofn.nFilterIndex =2;
        ofn.nMaxFile     =BUFFER_SIZE;
        ofn.Flags        =OFN_ALLOWMULTISELECT|OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_EXPLORER|OFN_ENABLESIZING;

        if (GetOpenFileNameW(&ofn))
        {
          wchar_t wszFile[MAX_PATH];
          wchar_t *wpFile=wszBuf + lstrlenW(wszBuf) + 1;
          SESSIONITEMW *lpItemElement;
          int nIndex;

          SendMessage(hWndItemsList, LB_SETSEL, FALSE, -1);

          if (*wpFile)
          {
            if (*(wpFile - 2) == '\\') *(wpFile - 2)='\0';

            do
            {
              if (IsPathFullW(wpFile))
                lstrcpyW(wszFile, wpFile);  //.lnk target
              else
                wsprintfW(wszFile, L"%s\\%s", wszBuf, wpFile);

              if (lpItemElement=AddSessionItemW(lpVirtualSessionW))
              {
                lstrcpynW(lpItemElement->wszItemFile, wszFile, MAX_PATH);
                nIndex=SendMessageW(hWndItemsList, LB_ADDSTRING, 0, (LPARAM)wszFile);
                SendMessage(hWndItemsList, LB_SETSEL, TRUE, nIndex);
                lpVirtualSessionW->bModified=TRUE;
              }
            }
            while (*(wpFile+=lstrlenW(wpFile) + 1));
          }
          else
          {
            if (lpItemElement=AddSessionItemW(lpVirtualSessionW))
            {
              lstrcpynW(lpItemElement->wszItemFile, wszBuf, MAX_PATH);
              nIndex=SendMessageW(hWndItemsList, LB_ADDSTRING, 0, (LPARAM)wszBuf);
              SendMessage(hWndItemsList, LB_SETSEL, TRUE, nIndex);
              lpVirtualSessionW->bModified=TRUE;
            }
          }
          PostMessage(hDlg, AKDLL_UPDATESAVEBUTTON, 0, 0);
        }
      }
    }
    else if (LOWORD(wParam) == IDC_ITEM_MOVEDOWN)
    {
      if (lpVirtualSessionW)
      {
        int *lpSelItems;
        int nSelCount;
        int nMaxIndex;
        int nOldIndex=-1;
        int nNewIndex=-1;
        int i;

        nMaxIndex=SendMessage(hWndItemsList, LB_GETCOUNT, 0, 0) - 1;

        if (nSelCount=GetListBoxSelItems(hWndItemsList, &lpSelItems))
        {
          for (i=--nSelCount; i >= 0; --i)
          {
            if (lpSelItems[i] < nMaxIndex)
            {
              if (nNewIndex == -1 && i < nSelCount)
              {
                if (lpSelItems[i] + 1 >= lpSelItems[i + 1])
                  continue;
              }
              nOldIndex=lpSelItems[i];
              nNewIndex=lpSelItems[i] + 1;
              MoveListBoxItemW(hWndItemsList, nOldIndex, nNewIndex);
              MoveSessionItemW(lpVirtualSessionW, nOldIndex + 1, nNewIndex + 1);
              SendMessage(hWndItemsList, LB_SETSEL, TRUE, nNewIndex);
              lpVirtualSessionW->bModified=TRUE;
            }
          }
          FreeListBoxSelItems(&lpSelItems);
        }
        PostMessage(hDlg, AKDLL_UPDATESAVEBUTTON, 0, 0);
      }
    }
    else if (LOWORD(wParam) == IDC_ITEM_MOVEUP)
    {
      if (lpVirtualSessionW)
      {
        int *lpSelItems;
        int nSelCount;
        int nMinIndex;
        int nOldIndex=-1;
        int nNewIndex=-1;
        int i;

        nMinIndex=0;

        if (nSelCount=GetListBoxSelItems(hWndItemsList, &lpSelItems))
        {
          for (i=0; i < nSelCount; ++i)
          {
            if (lpSelItems[i] > nMinIndex)
            {
              if (nNewIndex == -1 && i > 0)
              {
                if (lpSelItems[i] - 1 <= lpSelItems[i - 1])
                  continue;
              }
              nOldIndex=lpSelItems[i];
              nNewIndex=lpSelItems[i] - 1;
              MoveListBoxItemW(hWndItemsList, nOldIndex, nNewIndex);
              MoveSessionItemW(lpVirtualSessionW, nOldIndex + 1, nNewIndex + 1);
              SendMessage(hWndItemsList, LB_SETSEL, TRUE, nNewIndex);
              lpVirtualSessionW->bModified=TRUE;
            }
          }
          FreeListBoxSelItems(&lpSelItems);
        }
        PostMessage(hDlg, AKDLL_UPDATESAVEBUTTON, 0, 0);
      }
    }
    else if (LOWORD(wParam) == IDC_ITEM_DELETE)
    {
      if (lpVirtualSessionW)
      {
        int *lpSelItems;
        int nSelCount;
        int i;

        if (nSelCount=GetListBoxSelItems(hWndItemsList, &lpSelItems))
        {
          for (i=--nSelCount; i >= 0; --i)
          {
            SendMessage(hWndItemsList, LB_DELETESTRING, lpSelItems[i], 0);
            DeleteSessionItemW(lpVirtualSessionW, lpSelItems[i] + 1);
            lpVirtualSessionW->bModified=TRUE;
          }
          FreeListBoxSelItems(&lpSelItems);
        }
        PostMessage(hDlg, AKDLL_UPDATESAVEBUTTON, 0, 0);
      }
    }
    else if (LOWORD(wParam) == IDC_ITEM_DELETEOLD)
    {
      if (lpVirtualSessionW)
      {
        SESSIONITEMW *lpElement;
        int nMaxIndex;
        int i;

        nMaxIndex=SendMessage(hWndItemsList, LB_GETCOUNT, 0, 0) - 1;

        for (i=nMaxIndex; i >= 0; --i)
        {
          if (lpElement=GetSessionItemW(lpVirtualSessionW, i + 1))
          {
            if (!FileExistsW(lpElement->wszItemFile))
            {
              SendMessage(hWndItemsList, LB_DELETESTRING, i, 0);
              DeleteSessionItemW(lpVirtualSessionW, i + 1);
              lpVirtualSessionW->bModified=TRUE;
            }
          }
          else break;
        }
        PostMessage(hDlg, AKDLL_UPDATESAVEBUTTON, 0, 0);
      }
    }
    else if (LOWORD(wParam) == IDC_SETTINGS)
    {
      DialogBoxW(hInstanceDLL, MAKEINTRESOURCEW(IDD_SETTINGS), hDlg, (DLGPROC)SettingsDlgProcW);
    }
    else if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL)
    {
      if (SaveSessionPromptW() == IDCANCEL)
        return 0;
      if (lpVirtualSessionW)
        FreeSessionItemsW(lpVirtualSessionW);
      lpVirtualSessionW=NULL;
      lpRealSessionW=NULL;

      FreeSessionsW(&hSessionStack);
      DestroyMenu(hMenuList);

      if (!IsWindowEnabled(hMainWnd)) EnableWindow(hMainWnd, TRUE);
      DestroyWindow(hWndMainDlg);
      hWndMainDlg=NULL;
      CloseHandle(hThread);
      hThread=NULL;
      ExitThread(0);
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }

  if (DialogResizeMessages(&drs[0], &rcMainInitDialog, &rcMainCurrentDialog, hDlg, uMsg, wParam, lParam))
    bMainDialogRectSave=TRUE;

  return FALSE;
}

BOOL CALLBACK InputBoxDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndInputBox;
  static HWND hWndInputBoxEdit;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndInputBox=GetDlgItem(hDlg, IDC_INPUTBOX_EDIT);
    hWndInputBoxEdit=GetDlgItem(hWndInputBox, IDC_COMBOBOXEDIT);

    if (lParam == IDC_SESSION_SAVE)
      SetWindowTextA(hDlg, GetLangStringA(wLangSystem, STRID_SAVE));
    else if (lParam == IDC_SESSION_COPY)
      SetWindowTextA(hDlg, GetLangStringA(wLangSystem, STRID_COPY));
    else if (lParam == IDC_SESSION_RENAME)
      SetWindowTextA(hDlg, GetLangStringA(wLangSystem, STRID_RENAME));
    SetDlgItemTextA(hDlg, IDC_INPUTBOX_LABEL, GetLangStringA(wLangSystem, STRID_NEWNAME));
    SetDlgItemTextA(hDlg, IDOK, GetLangStringA(wLangSystem, STRID_OK));
    SetDlgItemTextA(hDlg, IDCANCEL, GetLangStringA(wLangSystem, STRID_CANCEL));

    SendMessage(hWndInputBoxEdit, EM_LIMITTEXT, MAX_PATH, 0);
    FillSessionsListA(&hSessionStack, hWndInputBox);
    if (lpVirtualSessionA)
    {
      if (lParam == IDC_SESSION_COPY)
      {
        char szSessionName[MAX_PATH];
        int nNumber=2;
        int i;

        lstrcpynA(szSessionName, lpVirtualSessionA->szSessionName, MAX_PATH);
        i=lstrlenA(szSessionName) - 1;

        if (szSessionName[i] == L')')
        {
          while (--i >= 0 && szSessionName[i] >= '0' && szSessionName[i] <= '9');
          if (szSessionName[i] == L'(')
          {
            nNumber=xatoiA(szSessionName + i + 1) + 1;
            szSessionName[i]=L'\0';
          }
        }
        for (i=nNumber; ; ++i)
        {
          wsprintfA(szBuf, "%s(%d)", szSessionName, i);
          if (!GetSessionA(&hSessionStack, szBuf))
          {
            SetWindowTextA(hWndInputBox, szBuf);
            break;
          }
        }
      }
      else SetWindowTextA(hWndInputBox, lpVirtualSessionA->szSessionName);
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_INPUTBOX_EDIT)
    {
      if (HIWORD(wParam) == CBN_DROPDOWN)
      {
        UpdateComboBoxDropWidthA(hWndInputBox);
      }
    }
    else if (LOWORD(wParam) == IDOK)
    {
      GetWindowTextA(hWndInputBox, szBuf, MAX_PATH);
      lstrcpynA(szDlgInputBox, szBuf, MAX_PATH);

      EndDialog(hDlg, 1);
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

BOOL CALLBACK InputBoxDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndInputBox;
  static HWND hWndInputBoxEdit;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndInputBox=GetDlgItem(hDlg, IDC_INPUTBOX_EDIT);
    hWndInputBoxEdit=GetDlgItem(hWndInputBox, IDC_COMBOBOXEDIT);

    if (lParam == IDC_SESSION_SAVE)
      SetWindowTextW(hDlg, GetLangStringW(wLangSystem, STRID_SAVE));
    else if (lParam == IDC_SESSION_COPY)
      SetWindowTextW(hDlg, GetLangStringW(wLangSystem, STRID_COPY));
    else if (lParam == IDC_SESSION_RENAME)
      SetWindowTextW(hDlg, GetLangStringW(wLangSystem, STRID_RENAME));
    SetDlgItemTextW(hDlg, IDC_INPUTBOX_LABEL, GetLangStringW(wLangSystem, STRID_NEWNAME));
    SetDlgItemTextW(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));
    SetDlgItemTextW(hDlg, IDCANCEL, GetLangStringW(wLangSystem, STRID_CANCEL));

    SendMessage(hWndInputBoxEdit, EM_LIMITTEXT, MAX_PATH, 0);
    FillSessionsListW(&hSessionStack, hWndInputBox);
    if (lpVirtualSessionW)
    {
      if (lParam == IDC_SESSION_COPY)
      {
        wchar_t wszSessionName[MAX_PATH];
        int nNumber=2;
        int i;

        lstrcpynW(wszSessionName, lpVirtualSessionW->wszSessionName, MAX_PATH);
        i=lstrlenW(wszSessionName) - 1;

        if (wszSessionName[i] == L')')
        {
          while (--i >= 0 && wszSessionName[i] >= '0' && wszSessionName[i] <= '9');
          if (wszSessionName[i] == L'(')
          {
            nNumber=xatoiW(wszSessionName + i + 1) + 1;
            wszSessionName[i]=L'\0';
          }
        }
        for (i=nNumber; ; ++i)
        {
          wsprintfW(wszBuf, L"%s(%d)", wszSessionName, i);
          if (!GetSessionW(&hSessionStack, wszBuf))
          {
            SetWindowTextW(hWndInputBox, wszBuf);
            break;
          }
        }
      }
      else SetWindowTextW(hWndInputBox, lpVirtualSessionW->wszSessionName);
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_INPUTBOX_EDIT)
    {
      if (HIWORD(wParam) == CBN_DROPDOWN)
      {
        UpdateComboBoxDropWidthW(hWndInputBox);
      }
    }
    else if (LOWORD(wParam) == IDOK)
    {
      GetWindowTextW(hWndInputBox, wszBuf, MAX_PATH);
      lstrcpynW(wszDlgInputBox, wszBuf, MAX_PATH);

      EndDialog(hDlg, 1);
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

BOOL CALLBACK SettingsDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndOpenOnStart;
  static HWND hWndOpenName;
  static HWND hWndSaveOnExit;
  static HWND hWndSaveName;
  static HWND hWndModalDialog;
  BOOL bState;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndOpenOnStart=GetDlgItem(hDlg, IDC_SETTINGS_OPENONSTART);
    hWndOpenName=GetDlgItem(hDlg, IDC_SETTINGS_OPENNAME);
    hWndSaveOnExit=GetDlgItem(hDlg, IDC_SETTINGS_SAVEONEXIT);
    hWndSaveName=GetDlgItem(hDlg, IDC_SETTINGS_SAVENAME);
    hWndModalDialog=GetDlgItem(hDlg, IDC_SETTINGS_MODAL);

    SetWindowTextA(hDlg, GetLangStringA(wLangSystem, STRID_SETTINGS));
    SetDlgItemTextA(hDlg, IDC_SETTINGS_GROUP_SESSION, GetLangStringA(wLangSystem, STRID_SESSION));
    SetDlgItemTextA(hDlg, IDC_SETTINGS_OPENONSTART, GetLangStringA(wLangSystem, STRID_OPENONSTART));
    SetDlgItemTextA(hDlg, IDC_SETTINGS_SAVEONEXIT, GetLangStringA(wLangSystem, STRID_SAVEONEXIT));
    SetDlgItemTextA(hDlg, IDC_SETTINGS_MODAL, GetLangStringA(wLangSystem, STRID_MODALDIALOG));
    SetDlgItemTextA(hDlg, IDOK, GetLangStringA(wLangSystem, STRID_OK));
    SetDlgItemTextA(hDlg, IDCANCEL, GetLangStringA(wLangSystem, STRID_CANCEL));

    SendMessage(hWndOpenName, EM_LIMITTEXT, MAX_PATH, 0);
    SendMessage(hWndSaveName, EM_LIMITTEXT, MAX_PATH, 0);
    SetWindowTextA(hWndOpenName, szOpenOnStart);
    SetWindowTextA(hWndSaveName, szSaveOnExit);
    if (bOpenOnStart) SendMessage(hWndOpenOnStart, BM_SETCHECK, BST_CHECKED, 0);
    if (bSaveOnExit) SendMessage(hWndSaveOnExit, BM_SETCHECK, BST_CHECKED, 0);
    if (bModalDialog) SendMessage(hWndModalDialog, BM_SETCHECK, BST_CHECKED, 0);
    PostMessage(hDlg, WM_COMMAND, IDC_SETTINGS_OPENONSTART, 0);
    PostMessage(hDlg, WM_COMMAND, IDC_SETTINGS_SAVEONEXIT, 0);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_SETTINGS_OPENONSTART)
    {
      bState=SendMessage(hWndOpenOnStart, BM_GETCHECK, 0, 0);
      EnableWindow(hWndOpenName, bState);
    }
    else if (LOWORD(wParam) == IDC_SETTINGS_SAVEONEXIT)
    {
      bState=SendMessage(hWndSaveOnExit, BM_GETCHECK, 0, 0);
      EnableWindow(hWndSaveName, bState);
    }
    else if (LOWORD(wParam) == IDOK)
    {
      GetWindowTextA(hWndOpenName, szOpenOnStart, MAX_PATH);
      GetWindowTextA(hWndSaveName, szSaveOnExit, MAX_PATH);
      bOpenOnStart=SendMessage(hWndOpenOnStart, BM_GETCHECK, 0, 0);
      bSaveOnExit=SendMessage(hWndSaveOnExit, BM_GETCHECK, 0, 0);
      bModalDialog=SendMessage(hWndModalDialog, BM_GETCHECK, 0, 0);
      EnableWindow(hMainWnd, !bModalDialog);
      SaveOptionsA(OF_MAIN);

      EndDialog(hDlg, 0);
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

BOOL CALLBACK SettingsDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndOpenOnStart;
  static HWND hWndOpenName;
  static HWND hWndSaveOnExit;
  static HWND hWndSaveName;
  static HWND hWndModalDialog;
  BOOL bState;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndOpenOnStart=GetDlgItem(hDlg, IDC_SETTINGS_OPENONSTART);
    hWndOpenName=GetDlgItem(hDlg, IDC_SETTINGS_OPENNAME);
    hWndSaveOnExit=GetDlgItem(hDlg, IDC_SETTINGS_SAVEONEXIT);
    hWndSaveName=GetDlgItem(hDlg, IDC_SETTINGS_SAVENAME);
    hWndModalDialog=GetDlgItem(hDlg, IDC_SETTINGS_MODAL);

    SetWindowTextW(hDlg, GetLangStringW(wLangSystem, STRID_SETTINGS));
    SetDlgItemTextW(hDlg, IDC_SETTINGS_GROUP_SESSION, GetLangStringW(wLangSystem, STRID_SESSION));
    SetDlgItemTextW(hDlg, IDC_SETTINGS_OPENONSTART, GetLangStringW(wLangSystem, STRID_OPENONSTART));
    SetDlgItemTextW(hDlg, IDC_SETTINGS_SAVEONEXIT, GetLangStringW(wLangSystem, STRID_SAVEONEXIT));
    SetDlgItemTextW(hDlg, IDC_SETTINGS_MODAL, GetLangStringW(wLangSystem, STRID_MODALDIALOG));
    SetDlgItemTextW(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));
    SetDlgItemTextW(hDlg, IDCANCEL, GetLangStringW(wLangSystem, STRID_CANCEL));

    SendMessage(hWndOpenName, EM_LIMITTEXT, MAX_PATH, 0);
    SendMessage(hWndSaveName, EM_LIMITTEXT, MAX_PATH, 0);
    SetWindowTextW(hWndOpenName, wszOpenOnStart);
    SetWindowTextW(hWndSaveName, wszSaveOnExit);
    if (bOpenOnStart) SendMessage(hWndOpenOnStart, BM_SETCHECK, BST_CHECKED, 0);
    if (bSaveOnExit) SendMessage(hWndSaveOnExit, BM_SETCHECK, BST_CHECKED, 0);
    if (bModalDialog) SendMessage(hWndModalDialog, BM_SETCHECK, BST_CHECKED, 0);
    PostMessage(hDlg, WM_COMMAND, IDC_SETTINGS_OPENONSTART, 0);
    PostMessage(hDlg, WM_COMMAND, IDC_SETTINGS_SAVEONEXIT, 0);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_SETTINGS_OPENONSTART)
    {
      bState=SendMessage(hWndOpenOnStart, BM_GETCHECK, 0, 0);
      EnableWindow(hWndOpenName, bState);
    }
    else if (LOWORD(wParam) == IDC_SETTINGS_SAVEONEXIT)
    {
      bState=SendMessage(hWndSaveOnExit, BM_GETCHECK, 0, 0);
      EnableWindow(hWndSaveName, bState);
    }
    else if (LOWORD(wParam) == IDOK)
    {
      GetWindowTextW(hWndOpenName, wszOpenOnStart, MAX_PATH);
      GetWindowTextW(hWndSaveName, wszSaveOnExit, MAX_PATH);
      bOpenOnStart=SendMessage(hWndOpenOnStart, BM_GETCHECK, 0, 0);
      bSaveOnExit=SendMessage(hWndSaveOnExit, BM_GETCHECK, 0, 0);
      bModalDialog=SendMessage(hWndModalDialog, BM_GETCHECK, 0, 0);
      EnableWindow(hMainWnd, !bModalDialog);
      SaveOptionsW(OF_MAIN);

      EndDialog(hDlg, 0);
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

LRESULT CALLBACK NewListBoxProcA(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_GETDLGCODE)
  {
    MSG *msg=(MSG *)lParam;

    if (msg)
    {
      if (msg->message == WM_KEYDOWN)
      {
        if (msg->wParam == VK_RETURN)
        {
          return DLGC_WANTALLKEYS;
        }
      }
    }
  }

  if (uMsg == WM_KEYDOWN)
  {
    if (wParam == VK_RETURN)
    {
      if (!(GetKeyState(VK_CONTROL) & 0x80) &&
          !(GetKeyState(VK_MENU) & 0x80) &&
          !(GetKeyState(VK_SHIFT) & 0x80))
      {
        PostMessage(GetParent(hWnd), WM_COMMAND, IDC_ITEM_OPEN, 0);
        return TRUE;
      }
    }
    else if (wParam == VK_INSERT)
    {
      if (!(GetKeyState(VK_CONTROL) & 0x80) &&
          !(GetKeyState(VK_MENU) & 0x80) &&
          !(GetKeyState(VK_SHIFT) & 0x80))
      {
        PostMessage(GetParent(hWnd), WM_COMMAND, IDC_ITEM_ADD, 0);
        return TRUE;
      }
    }
    else if (wParam == VK_DELETE)
    {
      if (!(GetKeyState(VK_CONTROL) & 0x80) &&
          !(GetKeyState(VK_MENU) & 0x80) &&
          !(GetKeyState(VK_SHIFT) & 0x80))
      {
        PostMessage(GetParent(hWnd), WM_COMMAND, IDC_ITEM_DELETE, 0);
        return TRUE;
      }
    }
  }
  else if (uMsg == WM_SYSKEYDOWN)
  {
    if (wParam == VK_DOWN)
    {
      if (!(GetKeyState(VK_CONTROL) & 0x80) &&
           (GetKeyState(VK_MENU) & 0x80) &&
          !(GetKeyState(VK_SHIFT) & 0x80))
      {
        PostMessage(GetParent(hWnd), WM_COMMAND, IDC_ITEM_MOVEDOWN, 0);
        return TRUE;
      }
    }
    else if (wParam == VK_UP)
    {
      if (!(GetKeyState(VK_CONTROL) & 0x80) &&
           (GetKeyState(VK_MENU) & 0x80) &&
          !(GetKeyState(VK_SHIFT) & 0x80))
      {
        PostMessage(GetParent(hWnd), WM_COMMAND, IDC_ITEM_MOVEUP, 0);
        return TRUE;
      }
    }
  }
  return CallWindowProcA(OldListBoxProc, hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK NewListBoxProcW(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_GETDLGCODE)
  {
    MSG *msg=(MSG *)lParam;

    if (msg)
    {
      if (msg->message == WM_KEYDOWN)
      {
        if (msg->wParam == VK_RETURN)
        {
          return DLGC_WANTALLKEYS;
        }
      }
    }
  }

  if (uMsg == WM_KEYDOWN)
  {
    if (wParam == VK_RETURN)
    {
      if (!(GetKeyState(VK_CONTROL) & 0x80) &&
          !(GetKeyState(VK_MENU) & 0x80) &&
          !(GetKeyState(VK_SHIFT) & 0x80))
      {
        PostMessage(GetParent(hWnd), WM_COMMAND, IDC_ITEM_OPEN, 0);
        return TRUE;
      }
    }
    else if (wParam == VK_INSERT)
    {
      if (!(GetKeyState(VK_CONTROL) & 0x80) &&
          !(GetKeyState(VK_MENU) & 0x80) &&
          !(GetKeyState(VK_SHIFT) & 0x80))
      {
        PostMessage(GetParent(hWnd), WM_COMMAND, IDC_ITEM_ADD, 0);
        return TRUE;
      }
    }
    else if (wParam == VK_DELETE)
    {
      if (!(GetKeyState(VK_CONTROL) & 0x80) &&
          !(GetKeyState(VK_MENU) & 0x80) &&
          !(GetKeyState(VK_SHIFT) & 0x80))
      {
        PostMessage(GetParent(hWnd), WM_COMMAND, IDC_ITEM_DELETE, 0);
        return TRUE;
      }
    }
  }
  else if (uMsg == WM_SYSKEYDOWN)
  {
    if (wParam == VK_DOWN)
    {
      if (!(GetKeyState(VK_CONTROL) & 0x80) &&
           (GetKeyState(VK_MENU) & 0x80) &&
          !(GetKeyState(VK_SHIFT) & 0x80))
      {
        PostMessage(GetParent(hWnd), WM_COMMAND, IDC_ITEM_MOVEDOWN, 0);
        return TRUE;
      }
    }
    else if (wParam == VK_UP)
    {
      if (!(GetKeyState(VK_CONTROL) & 0x80) &&
           (GetKeyState(VK_MENU) & 0x80) &&
          !(GetKeyState(VK_SHIFT) & 0x80))
      {
        PostMessage(GetParent(hWnd), WM_COMMAND, IDC_ITEM_MOVEUP, 0);
        return TRUE;
      }
    }
  }
  return CallWindowProcW(OldListBoxProc, hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK NewMainProcA(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == AKDN_MAIN_ONSTART_SHOW)
  {
    if (bOpenOnStart)
    {
      OpenSessionFileA(szOpenOnStart);
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    if (bSaveOnExit)
    {
      HSTACK hCloseStack={0};
      SESSIONA ss={0};
      SESSIONA *lpElement;

      if (lpElement=AddCurrentSessionA(&hCloseStack, szSaveOnExit))
      {
        SaveSessionFileA(lpElement);
        FreeSessionItemsA(lpElement);
      }
      else
      {
        lstrcpynA(ss.szSessionName, szSaveOnExit, MAX_PATH);
        SaveSessionFileA(&ss);
      }
    }
  }
  else if (uMsg == AKDN_MAIN_ONFINISH)
  {
    NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
    if (hWndMainDlg) SendMessage(hWndMainDlg, WM_COMMAND, IDCANCEL, 0);
    return FALSE;
  }

  //Call next procedure
  return NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK NewMainProcW(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == AKDN_MAIN_ONSTART_SHOW)
  {
    if (bOpenOnStart)
    {
      OpenSessionFileW(wszOpenOnStart);
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    if (bSaveOnExit)
    {
      HSTACK hCloseStack={0};
      SESSIONW ss={0};
      SESSIONW *lpElement;

      if (lpElement=AddCurrentSessionW(&hCloseStack, wszSaveOnExit))
      {
        SaveSessionFileW(lpElement);
        FreeSessionItemsW(lpElement);
      }
      else
      {
        lstrcpynW(ss.wszSessionName, wszSaveOnExit, MAX_PATH);
        SaveSessionFileW(&ss);
      }
    }
  }
  else if (uMsg == AKDN_MAIN_ONFINISH)
  {
    NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
    if (hWndMainDlg) SendMessage(hWndMainDlg, WM_COMMAND, IDCANCEL, 0);
    return FALSE;
  }

  //Call next procedure
  return NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

BOOL LoadSessionFileA(SESSIONA *ss)
{
  SESSIONITEMA *lpItemElement;
  char szSessionFile[MAX_PATH];
  char szDocumentFile[MAX_PATH];
  wchar_t *wszText;
  wchar_t *wpText;
  HANDLE hFile;
  CHARRANGE cr;
  DWORD dwFileSize;
  DWORD dwBytesRead;
  DWORD dwUnicodeLen;
  BOOL bWordWrap;
  BOOL bReadOnly;
  BOOL bOvertypeMode;
  int nCodePage;
  int nTabActive;
  int i;

  wsprintfA(szSessionFile, "%s\\%s.session", szSessionsDir, ss->szSessionName);
  if ((hFile=CreateFileA(szSessionFile, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE)
    return FALSE;
  if ((dwFileSize=GetFileSize(hFile, NULL)) != INVALID_FILE_SIZE)
  {
    if (dwFileSize >= 2)
    {
      if (wszText=wpText=(wchar_t *)GlobalAlloc(GPTR, dwFileSize + 2))
      {
        if (ReadFile(hFile, wszText, dwFileSize, &dwBytesRead, NULL))
        {
          dwUnicodeLen=dwBytesRead / sizeof(wchar_t);
          wszText[dwUnicodeLen++]='\0';
          if (wszText[0] == 0xFEFF)
          {
            ++wpText;
            --dwUnicodeLen;
          }

          while (*wpText)
          {
            //Get file name
            for (i=0; i < MAX_PATH && *wpText != '\r' && *wpText != '\0'; wszBuf[i++]=*wpText++);
            wszBuf[i]='\0';
            WideCharToMultiByte(CP_ACP, 0, wszBuf, i + 1, szDocumentFile, MAX_PATH, NULL, NULL);

            //Next line
            while (*wpText != '\r' && *wpText != '\0') ++wpText;
            if (*wpText == '\0') goto FreeText;
            if (*++wpText == '\n') ++wpText;

            //Get edit parameters
            nTabActive=0;
            nCodePage=0;
            cr.cpMin=0;
            cr.cpMax=0;
            bWordWrap=FALSE;
            bReadOnly=FALSE;
            bOvertypeMode=FALSE;

            while (*wpText != '\r' && *wpText != '\0')
            {
              if (*wpText == '/')
              {
                ++wpText;

                if (!memcmp(wpText, L"Active=", 14))
                {
                  wpText+=7;
                  if (xatoiW(wpText) == 1) nTabActive=1;
                }
                else if (!memcmp(wpText, L"Codepage=", 18))
                {
                  wpText+=9;
                  nCodePage=xatoiW(wpText);
                }
                else if (!memcmp(wpText, L"Selection=", 20))
                {
                  wpText+=10;
                  cr.cpMin=cr.cpMax=xatoiW(wpText);
                  while (*wpText >= '0' && *wpText <= '9') ++wpText;
                  if (*wpText == '-') cr.cpMax=xatoiW(++wpText);
                }
                else if (!memcmp(wpText, L"WordWrap=", 18))
                {
                  wpText+=9;
                  bWordWrap=xatoiW(wpText);
                }
                else if (!memcmp(wpText, L"ReadOnly=", 18))
                {
                  wpText+=9;
                  bReadOnly=xatoiW(wpText);
                }
                else if (!memcmp(wpText, L"Overtype=", 14))
                {
                  wpText+=7;
                  bOvertypeMode=xatoiW(wpText);
                }
                continue;
              }
              ++wpText;
            }
            if (szDocumentFile[0])
            {
              if (lpItemElement=AddSessionItemA(ss))
              {
                lstrcpynA(lpItemElement->szItemFile, szDocumentFile, MAX_PATH);
                lpItemElement->nTabActive=nTabActive;
                lpItemElement->nCodePage=nCodePage;
                lpItemElement->nSelStart=cr.cpMin;
                lpItemElement->nSelEnd=cr.cpMax;
                lpItemElement->bWordWrap=bWordWrap;
                lpItemElement->bReadOnly=bReadOnly;
                lpItemElement->bOvertypeMode=bOvertypeMode;
              }
            }

            while (*wpText == '\r' || *wpText == '\n') ++wpText;
          }
        }
        FreeText:
        GlobalFree((HGLOBAL)wszText);
      }
    }
  }
  CloseHandle(hFile);
  return TRUE;
}

BOOL LoadSessionFileW(SESSIONW *ss)
{
  SESSIONITEMW *lpItemElement;
  wchar_t wszSessionFile[MAX_PATH];
  wchar_t wszDocumentFile[MAX_PATH];
  wchar_t *wszText;
  wchar_t *wpText;
  HANDLE hFile;
  CHARRANGE cr;
  DWORD dwFileSize;
  DWORD dwBytesRead;
  DWORD dwUnicodeLen;
  BOOL bWordWrap;
  BOOL bReadOnly;
  BOOL bOvertypeMode;
  int nCodePage;
  int nTabActive;
  int i;

  wsprintfW(wszSessionFile, L"%s\\%s.session", wszSessionsDir, ss->wszSessionName);
  if ((hFile=CreateFileW(wszSessionFile, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE)
    return FALSE;
  if ((dwFileSize=GetFileSize(hFile, NULL)) != INVALID_FILE_SIZE)
  {
    if (dwFileSize >= 2)
    {
      if (wszText=wpText=(wchar_t *)GlobalAlloc(GPTR, dwFileSize + 2))
      {
        if (ReadFile(hFile, wszText, dwFileSize, &dwBytesRead, NULL))
        {
          dwUnicodeLen=dwBytesRead / sizeof(wchar_t);
          wszText[dwUnicodeLen++]='\0';
          if (wszText[0] == 0xFEFF)
          {
            ++wpText;
            --dwUnicodeLen;
          }

          while (*wpText)
          {
            //Get file name
            for (i=0; i < MAX_PATH && *wpText != '\r' && *wpText != '\0'; wszBuf[i++]=*wpText++);
            wszBuf[i]='\0';
            lstrcpynW(wszDocumentFile, wszBuf, MAX_PATH);

            //Next line
            while (*wpText != '\r' && *wpText != '\0') ++wpText;
            if (*wpText == '\0') goto FreeText;
            if (*++wpText == '\n') ++wpText;

            //Get edit parameters
            nTabActive=0;
            nCodePage=0;
            cr.cpMin=0;
            cr.cpMax=0;
            bWordWrap=FALSE;
            bReadOnly=FALSE;
            bOvertypeMode=FALSE;

            while (*wpText != '\r' && *wpText != '\0')
            {
              if (*wpText == '/')
              {
                ++wpText;

                if (!memcmp(wpText, L"Active=", 14))
                {
                  wpText+=7;
                  if (xatoiW(wpText) == 1) nTabActive=1;
                }
                else if (!memcmp(wpText, L"Codepage=", 18))
                {
                  wpText+=9;
                  nCodePage=xatoiW(wpText);
                }
                else if (!memcmp(wpText, L"Selection=", 20))
                {
                  wpText+=10;
                  cr.cpMin=cr.cpMax=xatoiW(wpText);
                  while (*wpText >= '0' && *wpText <= '9') ++wpText;
                  if (*wpText == '-') cr.cpMax=xatoiW(++wpText);
                }
                else if (!memcmp(wpText, L"WordWrap=", 18))
                {
                  wpText+=9;
                  bWordWrap=xatoiW(wpText);
                }
                else if (!memcmp(wpText, L"ReadOnly=", 18))
                {
                  wpText+=9;
                  bReadOnly=xatoiW(wpText);
                }
                else if (!memcmp(wpText, L"Overtype=", 14))
                {
                  wpText+=7;
                  bOvertypeMode=xatoiW(wpText);
                }
                continue;
              }
              ++wpText;
            }
            if (wszDocumentFile[0])
            {
              if (lpItemElement=AddSessionItemW(ss))
              {
                lstrcpynW(lpItemElement->wszItemFile, wszDocumentFile, MAX_PATH);
                lpItemElement->nTabActive=nTabActive;
                lpItemElement->nCodePage=nCodePage;
                lpItemElement->nSelStart=cr.cpMin;
                lpItemElement->nSelEnd=cr.cpMax;
                lpItemElement->bWordWrap=bWordWrap;
                lpItemElement->bReadOnly=bReadOnly;
                lpItemElement->bOvertypeMode=bOvertypeMode;
              }
            }

            while (*wpText == '\r' || *wpText == '\n') ++wpText;
          }
        }
        FreeText:
        GlobalFree((HGLOBAL)wszText);
      }
    }
  }
  CloseHandle(hFile);
  return TRUE;
}

BOOL RenameSessionFileA(char *pOldSessionName, char *pNewSessionName)
{
  char szOldSession[MAX_PATH];
  char szNewSession[MAX_PATH];

  wsprintfA(szOldSession, "%s\\%s.session", szSessionsDir, pOldSessionName);
  wsprintfA(szNewSession, "%s\\%s.session", szSessionsDir, pNewSessionName);
  return MoveFileA(szOldSession, szNewSession);
}

BOOL RenameSessionFileW(wchar_t *wpOldSessionName, wchar_t *wpNewSessionName)
{
  wchar_t wszOldSession[MAX_PATH];
  wchar_t wszNewSession[MAX_PATH];

  wsprintfW(wszOldSession, L"%s\\%s.session", wszSessionsDir, wpOldSessionName);
  wsprintfW(wszNewSession, L"%s\\%s.session", wszSessionsDir, wpNewSessionName);
  return MoveFileW(wszOldSession, wszNewSession);
}

void OpenSessionFileA(char *pSessionName)
{
  SESSIONA ss={0};

  lstrcpynA(ss.szSessionName, pSessionName, MAX_PATH);

  if (LoadSessionFileA(&ss))
  {
    OpenSessionA(&ss.hItemsStack);
    FreeSessionItemsA(&ss);
  }
}

void OpenSessionFileW(wchar_t *wpSessionName)
{
  SESSIONW ss={0};

  lstrcpynW(ss.wszSessionName, wpSessionName, MAX_PATH);

  if (LoadSessionFileW(&ss))
  {
    OpenSessionW(&ss.hItemsStack);
    FreeSessionItemsW(&ss);
  }
}

void SaveSessionFileA(SESSIONA *ss)
{
  SESSIONITEMA *lpItemElement=(SESSIONITEMA *)ss->hItemsStack.first;
  char szSessionFile[MAX_PATH];
  HANDLE hFile;
  DWORD dwBytesWritten;
  int nLineLen;

  CreateDirectoryA(szSessionsDir, NULL);

  wsprintfA(szSessionFile, "%s\\%s.session", szSessionsDir, ss->szSessionName);
  if ((hFile=CreateFileA(szSessionFile, GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE)
    return;
  if (WriteFile(hFile, "\xFF\xFE", 2, &dwBytesWritten, NULL))
  {
    while (lpItemElement)
    {
      nLineLen=wsprintfA(szBuf, "%s\r\n/Active=%d /Codepage=%d /Selection=%d-%d /WordWrap=%d /ReadOnly=%d /Overtype=%d\r\n\r\n",
                                lpItemElement->szItemFile, lpItemElement->nTabActive, lpItemElement->nCodePage, lpItemElement->nSelStart, lpItemElement->nSelEnd, lpItemElement->bWordWrap, lpItemElement->bReadOnly, lpItemElement->bOvertypeMode);
      MultiByteToWideChar(CP_ACP, 0, szBuf, nLineLen + 1, wszBuf, BUFFER_SIZE);

      if (!WriteFile(hFile, wszBuf, lstrlenW(wszBuf) * sizeof(wchar_t), &dwBytesWritten, NULL))
        break;

      lpItemElement=lpItemElement->next;
    }
  }
  CloseHandle(hFile);
}

void SaveSessionFileW(SESSIONW *ss)
{
  SESSIONITEMW *lpItemElement=(SESSIONITEMW *)ss->hItemsStack.first;
  wchar_t wszSessionFile[MAX_PATH];
  HANDLE hFile;
  DWORD dwBytesWritten;

  CreateDirectoryW(wszSessionsDir, NULL);

  wsprintfW(wszSessionFile, L"%s\\%s.session", wszSessionsDir, ss->wszSessionName);
  if ((hFile=CreateFileW(wszSessionFile, GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE)
    return;
  if (WriteFile(hFile, "\xFF\xFE", 2, &dwBytesWritten, NULL))
  {
    while (lpItemElement)
    {
      wsprintfW(wszBuf, L"%s\r\n/Active=%d /Codepage=%d /Selection=%d-%d /WordWrap=%d /ReadOnly=%d /Overtype=%d\r\n\r\n",
                          lpItemElement->wszItemFile, lpItemElement->nTabActive, lpItemElement->nCodePage, lpItemElement->nSelStart, lpItemElement->nSelEnd, lpItemElement->bWordWrap, lpItemElement->bReadOnly, lpItemElement->bOvertypeMode);

      if (!WriteFile(hFile, wszBuf, lstrlenW(wszBuf) * sizeof(wchar_t), &dwBytesWritten, NULL))
        break;

      lpItemElement=lpItemElement->next;
    }
  }
  CloseHandle(hFile);
}

BOOL DeleteSessionFileA(char *pSessionName)
{
  char szSession[MAX_PATH];

  wsprintfA(szSession, "%s\\%s.session", szSessionsDir, pSessionName);
  return DeleteFileA(szSession);
}

BOOL DeleteSessionFileW(wchar_t *wpSessionName)
{
  wchar_t wszSession[MAX_PATH];

  wsprintfW(wszSession, L"%s\\%s.session", wszSessionsDir, wpSessionName);
  return DeleteFileW(wszSession);
}

int SaveSessionPromptA()
{
  int nChoice=0;

  if (lpVirtualSessionA && lpVirtualSessionA->bModified)
  {
    wsprintfA(szBuf, GetLangStringA(wLangSystem, STRID_SESSION_CHANGED), lpVirtualSessionA->szSessionName);
    nChoice=MessageBoxA(hWndMainDlg, szBuf, "Sessions plugin", MB_YESNOCANCEL|MB_ICONEXCLAMATION);

    if (nChoice == IDYES)
    {
      FreeSessionItemsA(lpRealSessionA);
      lpVirtualSessionA->bModified=FALSE;
      CopySessionA(lpVirtualSessionA, lpRealSessionA);
      SaveSessionFileA(lpRealSessionA);
    }
  }
  return nChoice;
}

int SaveSessionPromptW()
{
  int nChoice=0;

  if (lpVirtualSessionW && lpVirtualSessionW->bModified)
  {
    wsprintfW(wszBuf, GetLangStringW(wLangSystem, STRID_SESSION_CHANGED), lpVirtualSessionW->wszSessionName);
    nChoice=MessageBoxW(hWndMainDlg, wszBuf, L"Sessions plugin", MB_YESNOCANCEL|MB_ICONEXCLAMATION);

    if (nChoice == IDYES)
    {
      FreeSessionItemsW(lpRealSessionW);
      lpVirtualSessionW->bModified=FALSE;
      CopySessionW(lpVirtualSessionW, lpRealSessionW);
      SaveSessionFileW(lpRealSessionW);
    }
  }
  return nChoice;
}

void FillSessionsA(HSTACK *hStack)
{
  SESSIONA *lpElement;
  WIN32_FIND_DATAA wfd;
  HANDLE hFind;
  char szFindFiles[MAX_PATH];
  char szBaseName[MAX_PATH];

  //Sessions files
  wsprintfA(szFindFiles, "%s\\*.session", szSessionsDir);

  if ((hFind=FindFirstFileA(szFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
  {
    do
    {
      if (wfd.cFileName[0] == '.' && (wfd.cFileName[1] == '\0' || (wfd.cFileName[1] == '.' && wfd.cFileName[2] == '\0'))) continue;

      GetBaseNameA(wfd.cFileName, szBaseName, MAX_PATH);

      if (lpElement=AddEmptySessionA(hStack, szBaseName))
      {
        LoadSessionFileA(lpElement);
      }
    }
    while (FindNextFileA(hFind, &wfd));

    FindClose(hFind);
  }
}

void FillSessionsW(HSTACK *hStack)
{
  SESSIONW *lpElement;
  WIN32_FIND_DATAW wfd;
  HANDLE hFind;
  wchar_t wszFindFiles[MAX_PATH];
  wchar_t wszBaseName[MAX_PATH];

  //Sessions files
  wsprintfW(wszFindFiles, L"%s\\*.session", wszSessionsDir);

  if ((hFind=FindFirstFileW(wszFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
  {
    do
    {
      if (wfd.cFileName[0] == '.' && (wfd.cFileName[1] == '\0' || (wfd.cFileName[1] == '.' && wfd.cFileName[2] == '\0'))) continue;

      GetBaseNameW(wfd.cFileName, wszBaseName, MAX_PATH);

      if (lpElement=AddEmptySessionW(hStack, wszBaseName))
      {
        LoadSessionFileW(lpElement);
      }
    }
    while (FindNextFileW(hFind, &wfd));

    FindClose(hFind);
  }
}

SESSIONA* AddEmptySessionA(HSTACK *hStack, char *pSessionName)
{
  SESSIONA *lpElement=NULL;

  if (!StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(SESSIONA)))
  {
    lstrcpynA(lpElement->szSessionName, pSessionName, MAX_PATH);
  }
  return lpElement;
}

SESSIONW* AddEmptySessionW(HSTACK *hStack, wchar_t *wpSessionName)
{
  SESSIONW *lpElement=NULL;

  if (!StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(SESSIONW)))
  {
    lstrcpynW(lpElement->wszSessionName, wpSessionName, MAX_PATH);
  }
  return lpElement;
}

SESSIONA* AddCurrentSessionA(HSTACK *hStack, char *pSessionName)
{
  SESSIONA *lpSessionElement=NULL;
  SESSIONITEMA *lpItemElement=NULL;
  EDITINFO ei;
  TCITEMA tcItem;
  CHARRANGE cr;
  HWND hWndFrameActive;
  int nTabActive;
  int nItem=0;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
  {
    hWndFrameActive=GetParent(ei.hWndEdit);

    if (lpSessionElement=AddEmptySessionA(hStack, pSessionName))
    {
      while (1)
      {
        tcItem.mask=TCIF_PARAM;
        if (!SendMessage(hTab, TCM_GETITEMA, nItem++, (LPARAM)&tcItem))
          break;

        if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)GetDlgItem((HWND)tcItem.lParam, ID_EDIT), (LPARAM)&ei))
        {
          if (*(char *)ei.pFile)
          {
            nTabActive=(hWndFrameActive == (HWND)tcItem.lParam)?1:0;
            SendMessage(ei.hWndEdit, EM_EXGETSEL, 0, (LPARAM)&cr);

            if (lpItemElement=AddSessionItemA(lpSessionElement))
            {
              lstrcpynA(lpItemElement->szItemFile, (char *)ei.pFile, MAX_PATH);
              lpItemElement->nTabActive=nTabActive;
              lpItemElement->nCodePage=ei.nCodePage;
              lpItemElement->nSelStart=cr.cpMin;
              lpItemElement->nSelEnd=cr.cpMax;
              lpItemElement->bWordWrap=ei.bWordWrap;
              lpItemElement->bReadOnly=ei.bReadOnly;
              lpItemElement->bOvertypeMode=ei.bOvertypeMode;
            }
          }
        }
      }
    }
  }
  return lpSessionElement;
}

SESSIONW* AddCurrentSessionW(HSTACK *hStack, wchar_t *wpSessionName)
{
  SESSIONW *lpSessionElement=NULL;
  SESSIONITEMW *lpItemElement=NULL;
  EDITINFO ei;
  TCITEMW tcItem;
  CHARRANGE cr;
  HWND hWndFrameActive;
  int nTabActive;
  int nItem=0;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
  {
    hWndFrameActive=GetParent(ei.hWndEdit);

    if (lpSessionElement=AddEmptySessionW(hStack, wpSessionName))
    {
      while (1)
      {
        tcItem.mask=TCIF_PARAM;
        if (!SendMessage(hTab, TCM_GETITEMW, nItem++, (LPARAM)&tcItem))
          break;

        if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)GetDlgItem((HWND)tcItem.lParam, ID_EDIT), (LPARAM)&ei))
        {
          if (*(wchar_t *)ei.pFile)
          {
            nTabActive=(hWndFrameActive == (HWND)tcItem.lParam)?1:0;
            SendMessage(ei.hWndEdit, EM_EXGETSEL, 0, (LPARAM)&cr);

            if (lpItemElement=AddSessionItemW(lpSessionElement))
            {
              lstrcpynW(lpItemElement->wszItemFile, (wchar_t *)ei.pFile, MAX_PATH);
              lpItemElement->nTabActive=nTabActive;
              lpItemElement->nCodePage=ei.nCodePage;
              lpItemElement->nSelStart=cr.cpMin;
              lpItemElement->nSelEnd=cr.cpMax;
              lpItemElement->bWordWrap=ei.bWordWrap;
              lpItemElement->bReadOnly=ei.bReadOnly;
              lpItemElement->bOvertypeMode=ei.bOvertypeMode;
            }
          }
        }
      }
    }
  }
  return lpSessionElement;
}

SESSIONA* GetSessionA(HSTACK *hStack, char *pSessionName)
{
  SESSIONA *lpElement=(SESSIONA *)hStack->first;

  while (lpElement)
  {
    if (!lstrcmpiA(lpElement->szSessionName, pSessionName))
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

SESSIONW* GetSessionW(HSTACK *hStack, wchar_t *wpSessionName)
{
  SESSIONW *lpElement=(SESSIONW *)hStack->first;

  while (lpElement)
  {
    if (!lstrcmpiW(lpElement->wszSessionName, wpSessionName))
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

void CopySessionA(SESSIONA *ssSource, SESSIONA *ssTarget)
{
  lstrcpynA(ssTarget->szSessionName, ssSource->szSessionName, MAX_PATH);
  ssTarget->hItemsStack.first=0;
  ssTarget->hItemsStack.last=0;
  ssTarget->bModified=ssSource->bModified;
  StackCopy((stack *)ssSource->hItemsStack.first, (stack *)ssSource->hItemsStack.last, (stack **)&ssTarget->hItemsStack.first, (stack **)&ssTarget->hItemsStack.last, sizeof(SESSIONITEMA));
}

void CopySessionW(SESSIONW *ssSource, SESSIONW *ssTarget)
{
  lstrcpynW(ssTarget->wszSessionName, ssSource->wszSessionName, MAX_PATH);
  ssTarget->hItemsStack.first=0;
  ssTarget->hItemsStack.last=0;
  ssTarget->bModified=ssSource->bModified;
  StackCopy((stack *)ssSource->hItemsStack.first, (stack *)ssSource->hItemsStack.last, (stack **)&ssTarget->hItemsStack.first, (stack **)&ssTarget->hItemsStack.last, sizeof(SESSIONITEMW));
}

void OpenSessionA(HSTACK *hStack)
{
  SESSIONITEMA *lpItemElement=(SESSIONITEMA *)hStack->first;
  EDITINFO ei;
  HWND hWndActivate=NULL;

  while (lpItemElement)
  {
    if (FileExistsA(lpItemElement->szItemFile))
    {
      OpenSessionItemA(lpItemElement);

      if (lpItemElement->nTabActive)
      {
        if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
        {
          hWndActivate=GetParent(ei.hWndEdit);
        }
      }
    }
    lpItemElement=lpItemElement->next;
  }

  if (hWndActivate) SendMessage(hMdiClient, WM_MDIACTIVATE, (WPARAM)hWndActivate, 0);
}

void OpenSessionW(HSTACK *hStack)
{
  SESSIONITEMW *lpItemElement=(SESSIONITEMW *)hStack->first;
  EDITINFO ei;
  HWND hWndActivate=NULL;

  while (lpItemElement)
  {
    if (FileExistsW(lpItemElement->wszItemFile))
    {
      OpenSessionItemW(lpItemElement);

      if (lpItemElement->nTabActive)
      {
        if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
        {
          hWndActivate=GetParent(ei.hWndEdit);
        }
      }
    }
    lpItemElement=lpItemElement->next;
  }

  if (hWndActivate) SendMessage(hMdiClient, WM_MDIACTIVATE, (WPARAM)hWndActivate, 0);
}

void CloseSessionA(HSTACK *hStack)
{
  SESSIONITEMA *lpItemElement=(SESSIONITEMA *)hStack->first;

  while (lpItemElement)
  {
    if (!CloseSessionItemA(lpItemElement)) break;

    lpItemElement=lpItemElement->next;
  }
}

void CloseSessionW(HSTACK *hStack)
{
  SESSIONITEMW *lpItemElement=(SESSIONITEMW *)hStack->first;

  while (lpItemElement)
  {
    if (!CloseSessionItemW(lpItemElement)) break;

    lpItemElement=lpItemElement->next;
  }
}

void DeleteSessionA(HSTACK *hStack, SESSIONA *ss)
{
  FreeSessionItemsA(ss);
  StackDelete((stack **)&hStack->first, (stack **)&hStack->last, (stack *)ss);
}

void DeleteSessionW(HSTACK *hStack, SESSIONW *ss)
{
  FreeSessionItemsW(ss);
  StackDelete((stack **)&hStack->first, (stack **)&hStack->last, (stack *)ss);
}

void FreeSessionsA(HSTACK *hStack)
{
  SESSIONA *lpElement=(SESSIONA *)hStack->first;

  while (lpElement)
  {
    FreeSessionItemsA(lpElement);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

void FreeSessionsW(HSTACK *hStack)
{
  SESSIONW *lpElement=(SESSIONW *)hStack->first;

  while (lpElement)
  {
    FreeSessionItemsW(lpElement);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}


SESSIONITEMA* AddSessionItemA(SESSIONA *ss)
{
  SESSIONITEMA *lpElement=NULL;

  StackInsertIndex((stack **)&ss->hItemsStack.first, (stack **)&ss->hItemsStack.last, (stack **)&lpElement, -1, sizeof(SESSIONITEMA));
  return lpElement;
}

SESSIONITEMW* AddSessionItemW(SESSIONW *ss)
{
  SESSIONITEMW *lpElement=NULL;

  StackInsertIndex((stack **)&ss->hItemsStack.first, (stack **)&ss->hItemsStack.last, (stack **)&lpElement, -1, sizeof(SESSIONITEMW));
  return lpElement;
}

SESSIONITEMA* GetSessionItemA(SESSIONA *ss, int nIndex)
{
  SESSIONITEMA *lpElement=NULL;

  StackGetElement((stack *)ss->hItemsStack.first, (stack *)ss->hItemsStack.last, (stack **)&lpElement, nIndex);
  return lpElement;
}

SESSIONITEMW* GetSessionItemW(SESSIONW *ss, int nIndex)
{
  SESSIONITEMW *lpElement=NULL;

  StackGetElement((stack *)ss->hItemsStack.first, (stack *)ss->hItemsStack.last, (stack **)&lpElement, nIndex);
  return lpElement;
}

void MoveSessionItemA(SESSIONA *ss, int nOldIndex, int nNewIndex)
{
  SESSIONITEMA *lpElement=NULL;

  if (lpElement=GetSessionItemA(ss, nOldIndex))
  {
    StackMoveIndex((stack **)&ss->hItemsStack.first, (stack **)&ss->hItemsStack.last, (stack *)lpElement, nNewIndex);
  }
}

void MoveSessionItemW(SESSIONW *ss, int nOldIndex, int nNewIndex)
{
  SESSIONITEMW *lpElement=NULL;

  if (lpElement=GetSessionItemW(ss, nOldIndex))
  {
    StackMoveIndex((stack **)&ss->hItemsStack.first, (stack **)&ss->hItemsStack.last, (stack *)lpElement, nNewIndex);
  }
}

void OpenSessionItemA(SESSIONITEMA *si)
{
  EDITINFO ei;
  OPENDOCUMENTA od;
  COPYDATASTRUCT cds;
  CHARRANGE cr;
  HWND hWndFrame=NULL;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
  {
    if (GetMenuState(hMainMenu, IDM_OPTIONS_SINGLEOPEN_FILE, MF_BYCOMMAND) & MF_CHECKED)
    {
      if (!lstrcmpiA(si->szItemFile, (char *)ei.pFile) || (hWndFrame=FindWindowExA(hMdiClient, NULL, APP_MDI_CLASSA, si->szItemFile)))
      {
        if (hWndFrame) SendMessage(hMdiClient, WM_MDIACTIVATE, (WPARAM)hWndFrame, 0);
        return;
      }
    }

    //Create empty document
    if (!ei.hWndEdit || ei.bModified || *(char *)ei.pFile)
      SendMessage(hMainWnd, WM_COMMAND, IDM_FILE_NEW, 0);
  }
  else SendMessage(hMainWnd, WM_COMMAND, IDM_FILE_NEW, 0);

  //Set edit parameters
  SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei);
  if (si->bWordWrap != ei.bWordWrap)
    SendMessage(hMainWnd, WM_COMMAND, IDM_VIEW_WORDWRAP, 0);
  if (si->bReadOnly != ei.bReadOnly)
    SendMessage(hMainWnd, WM_COMMAND, IDM_OPTIONS_READONLY, 0);
  if (si->bOvertypeMode != ei.bOvertypeMode)
    SendMessage(hMainWnd, WM_COMMAND, IDM_NONMENU_INSERTMODE, 0);

  //Open file
  od.hWnd=NULL;
  lstrcpyA(od.szFile, si->szItemFile);
  od.szWorkDir[0]='\0';
  od.dwFlags=OD_ADT_DETECT_BOM|(!si->nCodePage?OD_ADT_DETECT_CODEPAGE:0);
  od.nCodePage=si->nCodePage;
  od.bBOM=0;

  cds.dwData=CD_OPENDOCUMENT;
  cds.cbData=sizeof(OPENDOCUMENTA);
  cds.lpData=&od;
  SendMessage(hMainWnd, WM_COPYDATA, (WPARAM)hMainWnd, (LPARAM)&cds);

  //Set selection
  cr.cpMin=si->nSelStart;
  cr.cpMax=si->nSelEnd;
  SendMessage(ei.hWndEdit, EM_EXSETSEL, 0, (LPARAM)&cr);
}

void OpenSessionItemW(SESSIONITEMW *si)
{
  EDITINFO ei;
  OPENDOCUMENTW od;
  COPYDATASTRUCT cds;
  CHARRANGE cr;
  HWND hWndFrame=NULL;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
  {
    if (GetMenuState(hMainMenu, IDM_OPTIONS_SINGLEOPEN_FILE, MF_BYCOMMAND) & MF_CHECKED)
    {
      if (!lstrcmpiW(si->wszItemFile, (wchar_t *)ei.pFile) || (hWndFrame=FindWindowExW(hMdiClient, NULL, APP_MDI_CLASSW, si->wszItemFile)))
      {
        if (hWndFrame) SendMessage(hMdiClient, WM_MDIACTIVATE, (WPARAM)hWndFrame, 0);
        return;
      }
    }

    //Create empty document
    if (!ei.hWndEdit || ei.bModified || *(wchar_t *)ei.pFile)
      SendMessage(hMainWnd, WM_COMMAND, IDM_FILE_NEW, 0);
  }
  else SendMessage(hMainWnd, WM_COMMAND, IDM_FILE_NEW, 0);

  //Set edit parameters
  SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei);
  if (si->bWordWrap != ei.bWordWrap)
    SendMessage(hMainWnd, WM_COMMAND, IDM_VIEW_WORDWRAP, 0);
  if (si->bReadOnly != ei.bReadOnly)
    SendMessage(hMainWnd, WM_COMMAND, IDM_OPTIONS_READONLY, 0);
  if (si->bOvertypeMode != ei.bOvertypeMode)
    SendMessage(hMainWnd, WM_COMMAND, IDM_NONMENU_INSERTMODE, 0);

  //Open file
  od.hWnd=NULL;
  lstrcpyW(od.szFile, si->wszItemFile);
  od.szWorkDir[0]='\0';
  od.dwFlags=OD_ADT_DETECT_BOM|(!si->nCodePage?OD_ADT_DETECT_CODEPAGE:0);
  od.nCodePage=si->nCodePage;
  od.bBOM=0;

  cds.dwData=CD_OPENDOCUMENT;
  cds.cbData=sizeof(OPENDOCUMENTW);
  cds.lpData=&od;
  SendMessage(hMainWnd, WM_COPYDATA, (WPARAM)hMainWnd, (LPARAM)&cds);

  //Set selection
  cr.cpMin=si->nSelStart;
  cr.cpMax=si->nSelEnd;
  SendMessage(ei.hWndEdit, EM_EXSETSEL, 0, (LPARAM)&cr);
}

BOOL CloseSessionItemA(SESSIONITEMA *si)
{
  EDITINFO ei;
  HWND hWndFrame=NULL;
  BOOL bResult=TRUE;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
  {
    if (!lstrcmpiA(si->szItemFile, (char *)ei.pFile))
    {
      hWndFrame=GetParent(ei.hWndEdit);
    }
    else
    {
      if (hWndFrame=FindWindowExA(hMdiClient, NULL, APP_MDI_CLASSA, si->szItemFile))
      {
        SendMessage(hMdiClient, WM_MDIACTIVATE, (WPARAM)hWndFrame, 0);
        if (!SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei)) return FALSE;
      }
    }

    if (hWndFrame)
    {
      if (ei.bModified)
      {
        if (bModalDialog) EnableWindow(hMainWnd, TRUE);
        ShowWindow(hWndMainDlg, FALSE);
      }
      bResult=SendMessage(hMainWnd, WM_COMMAND, IDM_NONMENU_MDICLOSE, 0);
      if (ei.bModified)
      {
        ShowWindow(hWndMainDlg, TRUE);
        if (bModalDialog) EnableWindow(hMainWnd, FALSE);
      }
    }
  }
  return bResult;
}

BOOL CloseSessionItemW(SESSIONITEMW *si)
{
  EDITINFO ei;
  HWND hWndFrame=NULL;
  BOOL bResult=TRUE;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
  {
    if (!lstrcmpiW(si->wszItemFile, (wchar_t *)ei.pFile))
    {
      hWndFrame=GetParent(ei.hWndEdit);
    }
    else
    {
      if (hWndFrame=FindWindowExW(hMdiClient, NULL, APP_MDI_CLASSW, si->wszItemFile))
      {
        SendMessage(hMdiClient, WM_MDIACTIVATE, (WPARAM)hWndFrame, 0);
        if (!SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei)) return FALSE;
      }
    }

    if (hWndFrame)
    {
      if (ei.bModified)
      {
        if (bModalDialog) EnableWindow(hMainWnd, TRUE);
        ShowWindow(hWndMainDlg, FALSE);
      }
      bResult=SendMessage(hMainWnd, WM_COMMAND, IDM_NONMENU_MDICLOSE, 0);
      if (ei.bModified)
      {
        ShowWindow(hWndMainDlg, TRUE);
        if (bModalDialog) EnableWindow(hMainWnd, FALSE);
      }
    }
  }
  return bResult;
}

void DeleteSessionItemA(SESSIONA *ss, int nIndex)
{
  SESSIONITEMA *lpElement=NULL;

  if (!StackGetElement((stack *)ss->hItemsStack.first, (stack *)ss->hItemsStack.last, (stack **)&lpElement, nIndex))
  {
    StackDelete((stack **)&ss->hItemsStack.first, (stack **)&ss->hItemsStack.last, (stack *)lpElement);
  }
}

void DeleteSessionItemW(SESSIONW *ss, int nIndex)
{
  SESSIONITEMW *lpElement=NULL;

  if (!StackGetElement((stack *)ss->hItemsStack.first, (stack *)ss->hItemsStack.last, (stack **)&lpElement, nIndex))
  {
    StackDelete((stack **)&ss->hItemsStack.first, (stack **)&ss->hItemsStack.last, (stack *)lpElement);
  }
}

void FreeSessionItemsA(SESSIONA *ss)
{
  StackClear((stack **)&ss->hItemsStack.first, (stack **)&ss->hItemsStack.last);
}

void FreeSessionItemsW(SESSIONW *ss)
{
  StackClear((stack **)&ss->hItemsStack.first, (stack **)&ss->hItemsStack.last);
}

void FillSessionsListA(HSTACK *hStack, HWND hWnd)
{
  SESSIONA *lpElement=(SESSIONA *)hStack->first;

  while (lpElement)
  {
    SendMessageA(hWnd, CB_ADDSTRING, 0, (LPARAM)lpElement->szSessionName);

    lpElement=lpElement->next;
  }
}

void FillSessionsListW(HSTACK *hStack, HWND hWnd)
{
  SESSIONW *lpElement=(SESSIONW *)hStack->first;

  while (lpElement)
  {
    SendMessageW(hWnd, CB_ADDSTRING, 0, (LPARAM)lpElement->wszSessionName);

    lpElement=lpElement->next;
  }
}

void FillItemsListA(HSTACK *hStack, HWND hWnd)
{
  SESSIONITEMA *lpItemElement=(SESSIONITEMA *)hStack->first;

  while (lpItemElement)
  {
    SendMessageA(hWnd, LB_ADDSTRING, 0, (LPARAM)lpItemElement->szItemFile);

    lpItemElement=lpItemElement->next;
  }
}

void FillItemsListW(HSTACK *hStack, HWND hWnd)
{
  SESSIONITEMW *lpItemElement=(SESSIONITEMW *)hStack->first;

  while (lpItemElement)
  {
    SendMessageW(hWnd, LB_ADDSTRING, 0, (LPARAM)lpItemElement->wszItemFile);

    lpItemElement=lpItemElement->next;
  }
}

void FillItemsListCurrentA(HWND hWnd)
{
  EDITINFO ei;
  TCITEMA tcItem;
  int nItem=0;

  while (1)
  {
    tcItem.mask=TCIF_PARAM;
    if (!SendMessage(hTab, TCM_GETITEMA, nItem++, (LPARAM)&tcItem))
      break;

    if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)GetDlgItem((HWND)tcItem.lParam, ID_EDIT), (LPARAM)&ei))
    {
      if (*(char *)ei.pFile) SendMessageA(hWnd, LB_ADDSTRING, 0, (LPARAM)ei.pFile);
    }
  }
}

void FillItemsListCurrentW(HWND hWnd)
{
  EDITINFO ei;
  TCITEMA tcItem;
  int nItem=0;

  while (1)
  {
    tcItem.mask=TCIF_PARAM;
    if (!SendMessage(hTab, TCM_GETITEMW, nItem++, (LPARAM)&tcItem))
      break;

    if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)GetDlgItem((HWND)tcItem.lParam, ID_EDIT), (LPARAM)&ei))
    {
      if (*(wchar_t *)ei.pFile) SendMessageW(hWnd, LB_ADDSTRING, 0, (LPARAM)ei.pFile);
    }
  }
}

int UpdateListBoxHScrollA(HWND hWnd)
{
  char *pText;
  HDC hDC;
  HFONT hFont;
  HFONT hOldFont;
  SIZE size;
  int nTextLen;
  int nMaxWidth=0;
  int nIndex=0;

  if (hDC=GetDC(hWnd))
  {
    hFont=(HFONT)SendMessage(hWnd, WM_GETFONT, 0,0);
    hOldFont=(HFONT)SelectObject(hDC, hFont);

    while (1)
    {
      if ((nTextLen=SendMessage(hWnd, LB_GETTEXTLEN, nIndex, 0)) != LB_ERR)
      {
        if (pText=(char *)GlobalAlloc(GMEM_FIXED, nTextLen + 1))
        {
          SendMessageA(hWnd, LB_GETTEXT, nIndex, (LPARAM)pText);
          if (GetTextExtentPoint32A(hDC, pText, nTextLen, &size))
            nMaxWidth=max(size.cx, nMaxWidth);
          GlobalFree((HGLOBAL)pText);
        }
        else break;
      }
      else break;

      ++nIndex;
    }

    if (hOldFont) SelectObject(hDC, hOldFont);
    ReleaseDC(hWnd, hDC);
  }
  SendMessage(hWnd, LB_SETHORIZONTALEXTENT, nMaxWidth + 10, 0);
  return nMaxWidth;
}

int UpdateListBoxHScrollW(HWND hWnd)
{
  wchar_t *wpText;
  HDC hDC;
  HFONT hFont;
  HFONT hOldFont;
  SIZE size;
  int nTextLen;
  int nMaxWidth=0;
  int nIndex=0;

  if (hDC=GetDC(hWnd))
  {
    hFont=(HFONT)SendMessage(hWnd, WM_GETFONT, 0,0);
    hOldFont=(HFONT)SelectObject(hDC, hFont);

    while (1)
    {
      if ((nTextLen=SendMessage(hWnd, LB_GETTEXTLEN, nIndex, 0)) != LB_ERR)
      {
        if (wpText=(wchar_t *)GlobalAlloc(GMEM_FIXED, (nTextLen + 1) * sizeof(wchar_t)))
        {
          SendMessageW(hWnd, LB_GETTEXT, nIndex, (LPARAM)wpText);
          if (GetTextExtentPoint32W(hDC, wpText, nTextLen, &size))
            nMaxWidth=max(size.cx, nMaxWidth);
          GlobalFree((HGLOBAL)wpText);
        }
        else break;
      }
      else break;

      ++nIndex;
    }

    if (hOldFont) SelectObject(hDC, hOldFont);
    ReleaseDC(hWnd, hDC);
  }
  SendMessage(hWnd, LB_SETHORIZONTALEXTENT, nMaxWidth + 10, 0);
  return nMaxWidth;
}

int UpdateComboBoxDropWidthA(HWND hWnd)
{
  char *pText;
  HDC hDC;
  HFONT hFont;
  HFONT hOldFont;
  SIZE size;
  int nTextLen;
  int nMaxWidth=0;
  int nIndex=0;

  if (hDC=GetDC(hWnd))
  {
    hFont=(HFONT)SendMessage(hWnd, WM_GETFONT, 0,0);
    hOldFont=(HFONT)SelectObject(hDC, hFont);

    while (1)
    {
      if ((nTextLen=SendMessage(hWnd, CB_GETLBTEXTLEN, nIndex, 0)) != CB_ERR)
      {
        if (pText=(char *)GlobalAlloc(GMEM_FIXED, nTextLen + 1))
        {
          SendMessageA(hWnd, CB_GETLBTEXT, nIndex, (LPARAM)pText);
          if (GetTextExtentPoint32A(hDC, pText, nTextLen, &size))
            nMaxWidth=max(size.cx, nMaxWidth);
          GlobalFree((HGLOBAL)pText);
        }
        else break;
      }
      else break;

      ++nIndex;
    }

    if (hOldFont) SelectObject(hDC, hOldFont);
    ReleaseDC(hWnd, hDC);
  }
  SendMessage(hWnd, CB_SETDROPPEDWIDTH, nMaxWidth + 10, 0);
  return nMaxWidth;
}

int UpdateComboBoxDropWidthW(HWND hWnd)
{
  wchar_t *wpText;
  HDC hDC;
  HFONT hFont;
  HFONT hOldFont;
  SIZE size;
  int nTextLen;
  int nMaxWidth=0;
  int nIndex=0;

  if (hDC=GetDC(hWnd))
  {
    hFont=(HFONT)SendMessage(hWnd, WM_GETFONT, 0,0);
    hOldFont=(HFONT)SelectObject(hDC, hFont);

    while (1)
    {
      if ((nTextLen=SendMessage(hWnd, CB_GETLBTEXTLEN, nIndex, 0)) != CB_ERR)
      {
        if (wpText=(wchar_t *)GlobalAlloc(GMEM_FIXED, (nTextLen + 1) * sizeof(wchar_t)))
        {
          SendMessageW(hWnd, CB_GETLBTEXT, nIndex, (LPARAM)wpText);
          if (GetTextExtentPoint32W(hDC, wpText, nTextLen, &size))
            nMaxWidth=max(size.cx, nMaxWidth);
          GlobalFree((HGLOBAL)wpText);
        }
        else break;
      }
      else break;

      ++nIndex;
    }

    if (hOldFont) SelectObject(hDC, hOldFont);
    ReleaseDC(hWnd, hDC);
  }
  SendMessage(hWnd, CB_SETDROPPEDWIDTH, nMaxWidth + 10, 0);
  return nMaxWidth;
}

int MoveListBoxItemA(HWND hWnd, int nOldIndex, int nNewIndex)
{
  char *pText;
  int nIndex=LB_ERR;
  int nTextLen;

  if ((nTextLen=SendMessage(hWnd, LB_GETTEXTLEN, nOldIndex, 0)) != LB_ERR)
  {
    if (pText=(char *)GlobalAlloc(GMEM_FIXED, nTextLen + 1))
    {
      SendMessageA(hWnd, LB_GETTEXT, nOldIndex, (LPARAM)pText);
      SendMessage(hWnd, LB_DELETESTRING, nOldIndex, 0);
      nIndex=SendMessageA(hWnd, LB_INSERTSTRING, nNewIndex, (LPARAM)pText);
      GlobalFree((HGLOBAL)pText);
    }
  }
  return nIndex;
}

int MoveListBoxItemW(HWND hWnd, int nOldIndex, int nNewIndex)
{
  wchar_t *wpText;
  int nIndex=LB_ERR;
  int nTextLen;

  if ((nTextLen=SendMessage(hWnd, LB_GETTEXTLEN, nOldIndex, 0)) != LB_ERR)
  {
    if (wpText=(wchar_t *)GlobalAlloc(GMEM_FIXED, (nTextLen + 1) * sizeof(wchar_t)))
    {
      SendMessageW(hWnd, LB_GETTEXT, nOldIndex, (LPARAM)wpText);
      SendMessage(hWnd, LB_DELETESTRING, nOldIndex, 0);
      nIndex=SendMessageW(hWnd, LB_INSERTSTRING, nNewIndex, (LPARAM)wpText);
      GlobalFree((HGLOBAL)wpText);
    }
  }
  return nIndex;
}

int MoveComboBoxItemA(HWND hWnd, int nOldIndex, int nNewIndex)
{
  char *pText;
  int nIndex=CB_ERR;
  int nTextLen;

  if ((nTextLen=SendMessage(hWnd, CB_GETLBTEXTLEN, nOldIndex, 0)) != CB_ERR)
  {
    if (pText=(char *)GlobalAlloc(GMEM_FIXED, nTextLen + 1))
    {
      SendMessageA(hWnd, CB_GETLBTEXT, nOldIndex, (LPARAM)pText);
      SendMessage(hWnd, CB_DELETESTRING, nOldIndex, 0);
      nIndex=SendMessageA(hWnd, CB_INSERTSTRING, nNewIndex, (LPARAM)pText);
      GlobalFree((HGLOBAL)pText);
    }
  }
  return nIndex;
}

int MoveComboBoxItemW(HWND hWnd, int nOldIndex, int nNewIndex)
{
  wchar_t *wpText;
  int nIndex=CB_ERR;
  int nTextLen;

  if ((nTextLen=SendMessage(hWnd, CB_GETLBTEXTLEN, nOldIndex, 0)) != CB_ERR)
  {
    if (wpText=(wchar_t *)GlobalAlloc(GMEM_FIXED, (nTextLen + 1) * sizeof(wchar_t)))
    {
      SendMessageW(hWnd, CB_GETLBTEXT, nOldIndex, (LPARAM)wpText);
      SendMessage(hWnd, CB_DELETESTRING, nOldIndex, 0);
      nIndex=SendMessageW(hWnd, CB_INSERTSTRING, nNewIndex, (LPARAM)wpText);
      GlobalFree((HGLOBAL)wpText);
    }
  }
  return nIndex;
}

int GetListBoxSelItems(HWND hWnd, int **lpSelItems)
{
  int nSelCount;

  if (lpSelItems)
  {
    nSelCount=SendMessage(hWnd, LB_GETSELCOUNT, 0, 0);

    if (*lpSelItems=(int *)GlobalAlloc(GPTR, nSelCount * sizeof(int)))
    {
      return SendMessage(hWnd, LB_GETSELITEMS, nSelCount, (LPARAM)*lpSelItems);
    }
  }
  return 0;
}

void FreeListBoxSelItems(int **lpSelItems)
{
  if (lpSelItems && *lpSelItems)
  {
    GlobalFree((HGLOBAL)*lpSelItems);
    *lpSelItems=NULL;
  }
}

BOOL IsPathFullA(char *pPath)
{
  if (pPath[0] == '\\' && pPath[1] == '\\') return TRUE;
  if (pPath[0] != '\0' && pPath[1] == ':') return TRUE;
  return FALSE;
}

BOOL IsPathFullW(wchar_t *wpPath)
{
  if (wpPath[0] == '\\' && wpPath[1] == '\\') return TRUE;
  if (wpPath[0] != '\0' && wpPath[1] == ':') return TRUE;
  return FALSE;
}

int GetBaseNameA(char *pFile, char *szBaseName, int nBaseNameMaxLen)
{
  int nFileLen=lstrlenA(pFile);
  int nEndOffset=-1;
  int i;

  for (i=nFileLen - 1; i >= 0; --i)
  {
    if (pFile[i] == '\\')
      break;

    if (nEndOffset == -1)
    {
      if (pFile[i] == '.')
        nEndOffset=i;
    }
  }
  ++i;
  if (nEndOffset == -1) nEndOffset=nFileLen;
  nBaseNameMaxLen=min(nEndOffset - i + 1, nBaseNameMaxLen);
  lstrcpynA(szBaseName, pFile + i, nBaseNameMaxLen);

  return nBaseNameMaxLen;
}

int GetBaseNameW(wchar_t *wpFile, wchar_t *wszBaseName, int nBaseNameMaxLen)
{
  int nFileLen=lstrlenW(wpFile);
  int nEndOffset=-1;
  int i;

  for (i=nFileLen - 1; i >= 0; --i)
  {
    if (wpFile[i] == '\\')
      break;

    if (nEndOffset == -1)
    {
      if (wpFile[i] == '.')
        nEndOffset=i;
    }
  }
  ++i;
  if (nEndOffset == -1) nEndOffset=nFileLen;
  nBaseNameMaxLen=min(nEndOffset - i + 1, nBaseNameMaxLen);
  lstrcpynW(wszBaseName, wpFile + i, nBaseNameMaxLen);

  return nBaseNameMaxLen;
}

BOOL FileExistsA(char *pFile)
{
  WIN32_FIND_DATAA wfd;
  HANDLE hFind;

  if ((hFind=FindFirstFileA(pFile, &wfd)) == INVALID_HANDLE_VALUE)
    return FALSE;

  FindClose(hFind);
  return TRUE;
}

BOOL FileExistsW(wchar_t *wpFile)
{
  WIN32_FIND_DATAW wfd;
  HANDLE hFind;

  if ((hFind=FindFirstFileW(wpFile, &wfd)) == INVALID_HANDLE_VALUE)
    return FALSE;

  FindClose(hFind);
  return TRUE;
}

BOOL GetWindowPos(HWND hWnd, HWND hWndOwner, RECT *rc)
{
  if (GetWindowRect(hWnd, rc))
  {
    rc->right-=rc->left;
    rc->bottom-=rc->top;

    if (hWndOwner)
    {
      if (!ScreenToClient(hWndOwner, (POINT *)&rc->left))
        return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}

BOOL DialogResizeMessages(DIALOGRESIZE *drs, RECT *rcInit, RECT *rcCurrent, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_INITDIALOG)
  {
    RECT rcTemplate;

    GetWindowPos(hDlg, NULL, rcInit);
    rcTemplate=*rcCurrent;
    *rcCurrent=*rcInit;

    if (rcTemplate.right && rcTemplate.bottom)
    {
      rcTemplate.left=rcInit->left + (rcInit->right - rcTemplate.right) / 2;
      rcTemplate.top=rcInit->top + (rcInit->bottom - rcTemplate.bottom) / 2;
      SetWindowPos(hDlg, 0, rcTemplate.left, rcTemplate.top, rcTemplate.right, rcTemplate.bottom, SWP_NOZORDER);
    }
  }
  else if (uMsg == WM_GETMINMAXINFO)
  {
    MINMAXINFO *mmi=(MINMAXINFO *)lParam;

    mmi->ptMinTrackSize.x=rcInit->right;
    mmi->ptMinTrackSize.y=rcInit->bottom;
  }
  else if (uMsg == WM_SIZE)
  {
    if (lParam)
    {
      RECT rcTemplate;
      RECT rcControl;
      DWORD dwFlags;
      POINT pt;
      int i;

      GetWindowPos(hDlg, NULL, &rcTemplate);
      pt.x=rcTemplate.right - rcCurrent->right;
      pt.y=rcTemplate.bottom - rcCurrent->bottom;
      *rcCurrent=rcTemplate;

      for (i=0; drs[i].lpWnd; ++i)
      {
        if (*drs[i].lpWnd)
        {
          dwFlags=0;
          if (drs[i].dwType & DRS_SIZE)
            dwFlags|=SWP_NOMOVE;
          else if (drs[i].dwType & DRS_MOVE)
            dwFlags|=SWP_NOSIZE;
          GetWindowPos(*drs[i].lpWnd, hDlg, &rcControl);
          SetWindowPos(*drs[i].lpWnd, 0, rcControl.left + ((drs[i].dwType & DRS_X)?pt.x:0), rcControl.top + ((drs[i].dwType & DRS_Y)?pt.y:0), rcControl.right + ((drs[i].dwType & DRS_X)?pt.x:0), rcControl.bottom + ((drs[i].dwType & DRS_Y)?pt.y:0), dwFlags|SWP_NOZORDER);
        }
      }
      InvalidateRect(hDlg, NULL, TRUE);
      return TRUE;
    }
  }
  else if (uMsg == WM_PAINT)
  {
    PAINTSTRUCT ps;
    RECT rcGrip;
    HDC hDC;

    if (hDC=BeginPaint(hDlg, &ps))
    {
      GetClientRect(hDlg, &rcGrip);
      rcGrip.left=rcGrip.right - GetSystemMetrics(SM_CXVSCROLL);
      rcGrip.top=rcGrip.bottom - GetSystemMetrics(SM_CYVSCROLL);
      DrawFrameControl(hDC, &rcGrip, DFC_SCROLL, DFCS_SCROLLSIZEGRIP);
      EndPaint(hDlg, &ps);
    }
  }
  return FALSE;
}


//// Options

void ReadOptionsA()
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)szPluginName))
  {
    po.pOptionName="OpenOnStartEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bOpenOnStart;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="OpenOnStartSession";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)szOpenOnStart;
    po.dwData=MAX_PATH;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="SaveOnExitEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bSaveOnExit;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="SaveOnExitSession";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)szSaveOnExit;
    po.dwData=MAX_PATH;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="ModalDialogEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bModalDialog;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="WindowRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcMainCurrentDialog;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void ReadOptionsW()
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)wszPluginName))
  {
    po.pOptionName=L"OpenOnStartEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bOpenOnStart;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"OpenOnStartSession";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)wszOpenOnStart;
    po.dwData=MAX_PATH * sizeof(wchar_t);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"SaveOnExitEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bSaveOnExit;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"SaveOnExitSession";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)wszSaveOnExit;
    po.dwData=MAX_PATH * sizeof(wchar_t);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"ModalDialogEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bModalDialog;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"WindowRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcMainCurrentDialog;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsA(DWORD dwFlags)
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)szPluginName))
  {
    if (dwFlags & OF_MAIN)
    {
      po.pOptionName="OpenOnStartEnable";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bOpenOnStart;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="OpenOnStartSession";
      po.dwType=PO_STRING;
      po.lpData=(LPBYTE)szOpenOnStart;
      po.dwData=lstrlenA(szOpenOnStart) + 1;
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="SaveOnExitEnable";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bSaveOnExit;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="SaveOnExitSession";
      po.dwType=PO_STRING;
      po.lpData=(LPBYTE)szSaveOnExit;
      po.dwData=lstrlenA(szSaveOnExit) + 1;
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="ModalDialogEnable";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bModalDialog;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    if (dwFlags & OF_WINDOWRECT)
    {
      po.pOptionName="WindowRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&rcMainCurrentDialog;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }
    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsW(DWORD dwFlags)
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)wszPluginName))
  {
    if (dwFlags & OF_MAIN)
    {
      po.pOptionName=L"OpenOnStartEnable";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bOpenOnStart;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"OpenOnStartSession";
      po.dwType=PO_STRING;
      po.lpData=(LPBYTE)wszOpenOnStart;
      po.dwData=lstrlenW(wszOpenOnStart) * sizeof(wchar_t) + 2;
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"SaveOnExitEnable";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bSaveOnExit;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"SaveOnExitSession";
      po.dwType=PO_STRING;
      po.lpData=(LPBYTE)wszSaveOnExit;
      po.dwData=lstrlenW(wszSaveOnExit) * sizeof(wchar_t) + 2;
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"ModalDialogEnable";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bModalDialog;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    if (dwFlags & OF_WINDOWRECT)
    {
      po.pOptionName=L"WindowRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&rcMainCurrentDialog;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

char* GetLangStringA(LANGID wLangID, int nStringID)
{
  static char szStringBuf[MAX_PATH];

  WideCharToMultiByte(CP_ACP, 0, GetLangStringW(wLangID, nStringID), -1, szStringBuf, MAX_PATH, NULL, NULL);
  return szStringBuf;
}

wchar_t* GetLangStringW(LANGID wLangID, int nStringID)
{
  if (wLangID == LANG_RUSSIAN)
  {
    if (nStringID == STRID_SESSION)
      return L"\x0421\x0435\x0441\x0441\x0438\x044F";
    if (nStringID == STRID_OPEN)
      return L"\x041E\x0442\x043A\x0440\x044B\x0442\x044C";
    if (nStringID == STRID_SAVE)
      return L"\x0421\x043E\x0445\x0440\x0430\x043D\x0438\x0442\x044C";
    if (nStringID == STRID_COPY)
      return L"\x0421\x043A\x043E\x043F\x0438\x0440\x043E\x0432\x0430\x0442\x044C";
    if (nStringID == STRID_RENAME)
      return L"\x041F\x0435\x0440\x0435\x0438\x043C\x0435\x043D\x043E\x0432\x0430\x0442\x044C";
    if (nStringID == STRID_DELETE)
      return L"\x0423\x0434\x0430\x043B\x0438\x0442\x044C";
    if (nStringID == STRID_MENU_OPEN)
      return L"\x041E\x0442\x043A\x0440\x044B\x0442\x044C\tEnter";
    if (nStringID == STRID_MENU_ACTIVATE)
      return L"\x0410\x043A\x0442\x0438\x0432\x0438\x0440\x043E\x0432\x0430\x0442\x044C\tEnter";
    if (nStringID == STRID_MENU_ADD)
      return L"\x0414\x043E\x0431\x0430\x0432\x0438\x0442\x044C\x002E\x002E\x002E\tInsert";
    if (nStringID == STRID_MENU_MOVEDOWN)
      return L"\x041F\x0435\x0440\x0435\x043C\x0435\x0441\x0442\x0438\x0442\x044C\x0020\x0432\x043D\x0438\x0437\tAlt+Down";
    if (nStringID == STRID_MENU_MOVEUP)
      return L"\x041F\x0435\x0440\x0435\x043C\x0435\x0441\x0442\x0438\x0442\x044C\x0020\x0432\x0432\x0435\x0440\x0445\tAlt+Up";
    if (nStringID == STRID_MENU_DELETE)
      return L"\x0423\x0434\x0430\x043B\x0438\x0442\x044C\tDelete";
    if (nStringID == STRID_MENU_DELETEOLD)
      return L"\x0423\x0434\x0430\x043B\x0438\x0442\x044C\x0020\x043D\x0435\x0441\x0443\x0449\x0435\x0441\x0442\x0432\x0443\x044E\x0449\x0438\x0435";
    if (nStringID == STRID_NEWNAME)
      return L"\x041D\x043E\x0432\x043E\x0435\x0020\x0438\x043C\x044F\x003A";
    if (nStringID == STRID_SETTINGS)
      return L"\x041D\x0430\x0441\x0442\x0440\x043E\x0439\x043A\x0438\x002E\x002E\x002E";
    if (nStringID == STRID_OPENONSTART)
      return L"\x041E\x0442\x043A\x0440\x044B\x0442\x044C\x0020\x043F\x0440\x0438\x0020\x0437\x0430\x043F\x0443\x0441\x043A\x0435\x003A";
    if (nStringID == STRID_SAVEONEXIT)
      return L"\x0421\x043E\x0445\x0440\x0430\x043D\x0438\x0442\x044C\x0020\x043D\x0430\x0020\x0432\x044B\x0445\x043E\x0434\x0435\x003A";
    if (nStringID == STRID_MODALDIALOG)
      return L"\x041C\x043E\x0434\x0430\x043B\x044C\x043D\x044B\x0439\x0020\x0434\x0438\x0430\x043B\x043E\x0433";
    if (nStringID == STRID_OK)
      return L"\x004F\x004B";
    if (nStringID == STRID_CANCEL)
      return L"\x041E\x0442\x043C\x0435\x043D\x0430";
    if (nStringID == STRID_CLOSE)
      return L"\x0417\x0430\x043A\x0440\x044B\x0442\x044C";
    if (nStringID == STRID_CURRENTSESSION)
      return L"\x003C\x0422\x0435\x043A\x0443\x0449\x0430\x044F\x003E";
    if (nStringID == STRID_ALREADY_EXIST)
      return L"\x0421\x0435\x0441\x0441\x0438\x044F\x0020\x0022\x0025\x0073\x0022\x0020\x0443\x0436\x0435\x0020\x0441\x0443\x0449\x0435\x0441\x0442\x0432\x0443\x0435\x0442\x002E\x0020\x041F\x0435\x0440\x0435\x0437\x0430\x043F\x0438\x0441\x0430\x0442\x044C\x0020\x0435\x0451\x003F";
    if (nStringID == STRID_RENAME_ERROR)
      return L"\x041E\x0448\x0438\x0431\x043A\x0430\x0020\x043F\x0440\x0438\x0020\x043F\x0435\x0440\x0435\x0438\x043C\x0435\x043D\x043E\x0432\x0430\x043D\x0438\x0438\x0020\x0441\x0435\x0441\x0441\x0438\x0438\x002E";
    if (nStringID == STRID_SESSION_CHANGED)
      return L"\x0421\x0435\x0441\x0441\x0438\x044F\x0020\x0022\x0025\x0073\x0022\x0020\x0431\x044B\x043B\x0430\x0020\x0438\x0437\x043C\x0435\x043D\x0451\x043D\x0430\x002E\x0020\x0421\x043E\x0445\x0440\x0430\x043D\x0438\x0442\x044C\x003F";
    if (nStringID == STRID_CONFIRM_DELETE)
      return L"\x0412\x044B\x0020\x0443\x0432\x0435\x0440\x0435\x043D\x044B\x002C\x0020\x0447\x0442\x043E\x0020\x0445\x043E\x0442\x0438\x0442\x0435\x0020\x0443\x0434\x0430\x043B\x0438\x0442\x044C\x0020\x0441\x0435\x0441\x0441\x0438\x044E\x003F";
    if (nStringID == STRID_SDI_ISNTSUPPORTED)
      return L"\x0053\x0044\x0049\x0020\x0440\x0435\x0436\x0438\x043C\x0020\x043D\x0435\x0020\x043F\x043E\x0434\x0434\x0435\x0440\x0436\x0438\x0432\x0430\x0435\x0442\x0441\x044F\x002E";
    if (nStringID == STRID_FILTER)
      return L"\x0422\x0435\x043A\x0441\x0442\x043E\x0432\x044B\x0435\x0020\x0444\x0430\x0439\x043B\x044B (*.txt)\0*.txt\0\x0412\x0441\x0435\x0020\x0444\x0430\x0439\x043B\x044B (*.*)\0*.*\0\0";
  }
  else
  {
    if (nStringID == STRID_SESSION)
      return L"Session";
    if (nStringID == STRID_OPEN)
      return L"Open";
    if (nStringID == STRID_SAVE)
      return L"Save";
    if (nStringID == STRID_COPY)
      return L"Copy";
    if (nStringID == STRID_RENAME)
      return L"Rename";
    if (nStringID == STRID_DELETE)
      return L"Delete";
    if (nStringID == STRID_MENU_OPEN)
      return L"Open\tEnter";
    if (nStringID == STRID_MENU_ACTIVATE)
      return L"Activate\tEnter";
    if (nStringID == STRID_MENU_ADD)
      return L"Add...\tInsert";
    if (nStringID == STRID_MENU_MOVEDOWN)
      return L"Move down\tAlt+Down";
    if (nStringID == STRID_MENU_MOVEUP)
      return L"Move up\tAlt+Up";
    if (nStringID == STRID_MENU_DELETE)
      return L"Delete\tDelete";
    if (nStringID == STRID_MENU_DELETEOLD)
      return L"Delete non-existent";
    if (nStringID == STRID_NEWNAME)
      return L"New name";
    if (nStringID == STRID_SETTINGS)
      return L"Settings...";
    if (nStringID == STRID_OPENONSTART)
      return L"Open on start:";
    if (nStringID == STRID_SAVEONEXIT)
      return L"Save on exit:";
    if (nStringID == STRID_MODALDIALOG)
      return L"Modal dialog";
    if (nStringID == STRID_OK)
      return L"OK";
    if (nStringID == STRID_CANCEL)
      return L"Cancel";
    if (nStringID == STRID_CLOSE)
      return L"Close";
    if (nStringID == STRID_CURRENTSESSION)
      return L"<Current>";
    if (nStringID == STRID_ALREADY_EXIST)
      return L"Session \"%s\" already exists. Overwrite it?";
    if (nStringID == STRID_RENAME_ERROR)
      return L"Can't rename session.";
    if (nStringID == STRID_SESSION_CHANGED)
      return L"Session \"%s\" was modified. Do you want to save it?";
    if (nStringID == STRID_CONFIRM_DELETE)
      return L"Are you sure you want to delete session?";
    if (nStringID == STRID_SDI_ISNTSUPPORTED)
      return L"SDI mode is not supported.";
    if (nStringID == STRID_FILTER)
      return L"Text files (*.txt)\0*.txt\0All Files (*.*)\0*.*\0\0";
  }
  return L"";
}

BOOL IsExtCallParamValid(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return TRUE;
  return FALSE;
}

int GetExtCallParam(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return *(((int *)lParam) + nIndex);
  return 0;
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hInstanceDLL=pd->hInstanceDLL;
  hMainWnd=pd->hMainWnd;
  hMdiClient=pd->hMdiClient;
  hTab=pd->hTab;
  hMainMenu=pd->hMainMenu;
  hMainIcon=pd->hMainIcon;
  bOldWindows=pd->bOldWindows;
  wLangSystem=PRIMARYLANGID(pd->wLangSystem);

  if (bOldWindows)
  {
    int i;

    for (i=0; ((char *)pd->pFunction)[i] != ':'; ++i)
      szPluginName[i]=((char *)pd->pFunction)[i];
    szPluginName[i]='\0';
  }
  else
  {
    int i;

    for (i=0; ((wchar_t *)pd->pFunction)[i] != ':'; ++i)
      wszPluginName[i]=((wchar_t *)pd->pFunction)[i];
    wszPluginName[i]='\0';
  }

  if (bOldWindows)
  {
    wsprintfA(szSessionsDir, "%s\\AkelFiles\\Plugs\\Sessions", (char *)pd->pAkelDir);
    ReadOptionsA();
  }
  else
  {
    wsprintfW(wszSessionsDir, L"%s\\AkelFiles\\Plugs\\Sessions", (wchar_t *)pd->pAkelDir);
    ReadOptionsW();
  }
}

void InitMain()
{
  bInitMain=TRUE;

  //SubClass
  if (bOldWindows)
  {
    NewMainProcData=NULL;
    SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NewMainProcA, (LPARAM)&NewMainProcData);
  }
  else
  {
    NewMainProcData=NULL;
    SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NewMainProcW, (LPARAM)&NewMainProcData);
  }
}

void UninitMain()
{
  bInitMain=FALSE;

  //Remove subclass
  if (NewMainProcData)
  {
    SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NULL, (LPARAM)&NewMainProcData);
    NewMainProcData=NULL;
  }

  if (bOldWindows)
  {
    if (bMainDialogRectSave) SaveOptionsA(OF_WINDOWRECT);
    FreeSessionsA(&hSessionStack);
  }
  else
  {
    if (bMainDialogRectSave) SaveOptionsW(OF_WINDOWRECT);
    FreeSessionsW(&hSessionStack);
  }
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
    if (bInitMain) UninitMain();
  }
  return TRUE;
}
